import unittest

from cort.core.mentions import Mention
from cort.core.spans import Span
from cort.coreference.multigraph import features


__author__ = 'smartschat'


class TestFeatures(unittest.TestCase):
    def test_non_pronominal_string_match(self):
        self.assertEqual(
            True,
            features.non_pronominal_string_match(
                Mention(
                    None,
                    Span(0, 4),
                    {"tokens": ["the", "newly-elect", "leader", "'s", "wife"],
                     "pos": ["DT", "JJ", "NN", "POS", "NN"], "type": "NOM"}),
                Mention(
                    None,
                    Span(5, 7),
                    {"tokens": ["newly-elect", "leader", "wife"],
                     "pos": ["JJ", "NN", "NN"], "type": "NOM"})))
        self.assertEqual(
            False,
            features.non_pronominal_string_match(
                Mention(
                    None,
                    Span(0, 4),
                    {"tokens": ["the", "newly-elect", "leader", "'s", "wife"],
                     "pos": ["DT", "JJ", "NN", "POS", "NN"], "type": "NOM"}),
                Mention(
                    None,
                    Span(5, 5),
                    {"tokens": ["leader"], "pos": ["NN"], "type": "NOM"})))
        self.assertEqual(
            True,
            features.non_pronominal_string_match(
                Mention(
                    None,
                    Span(0, 0),
                    {"tokens": ["President"], "pos": ["NNP"], "type": "NAM"}),
                Mention(
                    None,
                    Span(1, 2),
                    {"tokens": ["the", "president"], "pos": ["DT", "NN"],
                     "type": "NOM"})))
        self.assertEqual(
            False,
            features.non_pronominal_string_match(
                Mention(
                    None,
                    Span(0, 0),
                    {"tokens": ["it"], "pos": ["PRP"], "type": "PRO"}),
                Mention(
                    None,
                    Span(1, 1),
                    {"tokens": ["IT"], "pos": ["NNP"], "type": "NAM"})))

    def test_head_match(self):
        self.assertEqual(
            True,
            features.head_match(
                Mention(
                    None,
                    Span(0, 4),
                    {"tokens": ["the", "newly-elect", "leader", "'s", "wife"],
                     "head": ["wife"], "type": "NOM",
                     "semantic_class": "PERSON"}),
                Mention(
                    None,
                    Span(5, 6),
                    {"tokens": ["the", "wife"], "head": ["wife"],
                     "type": "NOM", "semantic_class": "PERSON"})))
        self.assertEqual(
            False,
            features.head_match(
                Mention(
                    None,
                    Span(0, 4),
                    {"tokens": ["the", "newly-elect", "leader", "'s", "wife"],
                     "head": ["wife"], "type": "NOM",
                     "semantic_class": "PERSON"}),
                Mention(
                    None,
                    Span(5, 5),
                    {"tokens": ["leader"], "head": ["leader"], "type": "NOM",
                     "semantic_class": "PERSON"})))
        self.assertEqual(
            True,
            features.head_match(
                Mention(
                    None,
                    Span(0, 0),
                    {"tokens": ["President"], "head": ["President"],
                     "type": "NAM", "semantic_class": "PERSON"}),
                Mention(
                    None,
                    Span(1, 2),
                    {"tokens": ["the", "president"], "head": ["president"],
                     "type": "NOM", "semantic_class": "PERSON"})))
        self.assertEqual(
            False,
            features.head_match(
                Mention(
                    None,
                    Span(0, 0),
                    {"tokens": ["it"], "head": ["it"], "type": "PRO",
                     "semantic_class": "OBJECT"}),
                Mention(
                    None,
                    Span(1, 1),
                    {"tokens": ["it"], "head": ["it"], "type": "PRO",
                     "semantic_class": "OBJECT"})))
        self.assertEqual(
            False,
            features.head_match(
                Mention(
                    None,
                    Span(0, 1),
                    {"tokens": ["10", "percent"], "head": ["percent"],
                     "type": "NOM", "semantic_class": "NUMERIC"}),
                Mention(
                    None,
                    Span(2, 3),
                    {"tokens": ["Some", "percent"], "head": ["percent"],
                     "type": "NOM", "semantic_class": "NUMERIC"})))

    def test_pronoun_same_canonical_form(self):
        self.assertEqual(
            True,
            features.pronoun_same_canonical_form(
                Mention(
                    None,
                    Span(0, 0),
                    {"tokens": ["he"], "type": "PRO", "citation_form": "he"}),
                Mention(
                    None,
                    Span(1, 1),
                    {"tokens": ["he"], "type": "PRO", "citation_form": "he"})))

        self.assertEqual(
            True,
            features.pronoun_same_canonical_form(
                Mention(
                    None,
                    Span(0, 0),
                    {"tokens": ["he"], "type": "PRO", "citation_form": "he"}),
                Mention(
                    None,
                    Span(1, 1),
                    {"tokens": ["him"], "type": "PRO",
                     "citation_form": "he"})))

        self.assertEqual(
            False,
            features.pronoun_same_canonical_form(
                Mention(
                    None,
                    Span(0, 0),
                    {"tokens": ["US"], "type": "NAM"}),
                Mention(
                    None,
                    Span(1, 1),
                    {"tokens": ["us"], "type": "PRO"})))

    def test_speaker(self):
        self.assertEqual(True, features.speaker(
            Mention(None, Span(3, 3), {"tokens": ["you"], "type": "PRO", "citation_form": "you", "speaker": "snafu"}),
            Mention(None, Span(0, 0), {"tokens": ["I"], "type": "PRO", "citation_form": "i", "speaker": "foo"}),
        ))

        self.assertEqual(True, features.speaker(
            Mention(None, Span(3, 3), {"tokens": ["me"], "type": "PRO", "citation_form": "i", "speaker": "snafu"}),
            Mention(None, Span(0, 0), {"tokens": ["I"], "type": "PRO", "citation_form": "i", "speaker": "snafu"}),
        ))

        self.assertEqual(False, features.speaker(
            Mention(None, Span(3, 3), {"tokens": ["you"], "type": "PRO", "citation_form": "you", "speaker": "snafu"}),
            Mention(None, Span(0, 0), {"tokens": ["I"], "type": "PRO", "citation_form": "i", "speaker": "snafu"}),
        ))

        self.assertEqual(True, features.speaker(
            Mention(None, Span(3, 3), {"tokens": ["me"], "type": "PRO", "citation_form": "i", "speaker": "snafu"}),
            Mention(None, Span(0, 0), {"tokens": ["snafu"], "type": "NAM", "speaker": "-", "head": ["snafu"]}),
        ))

        self.assertEqual(False, features.speaker(
            Mention(None, Span(3, 3), {"tokens": ["me"], "type": "PRO", "citation_form": "i", "speaker": "snafu"}),
            Mention(None, Span(0, 0), {"tokens": ["foo"], "type": "NAM", "speaker": "-", "head": ["foo"]}),
        ))

    def test_not_pronoun_distance(self):
        self.assertEqual(False, features.not_pronoun_distance(
            Mention(None, Span(0, 0), {"tokens": ["he"], "type": "PRO", "citation_form": "he", "sentence_id": 10}),
            Mention(None, Span(100, 100), {"tokens": ["he"], "type": "PRO", "citation_form": "he", "sentence_id": 0})
        ))

        self.assertEqual(True, features.not_pronoun_distance(
            Mention(None, Span(100, 100), {"tokens": ["it"], "type": "PRO", "citation_form": "it", "sentence_id": 10}),
            Mention(None, Span(0, 0), {"tokens": ["company"], "type": "NOM", "sentence_id": 0})
        ))

        self.assertEqual(False, features.not_pronoun_distance(
            Mention(None, Span(100, 100), {"tokens": ["them"], "type": "PRO", "citation_form": "they", "sentence_id": 10}),
            Mention(None, Span(0, 0), {"tokens": ["company"], "type": "NOM", "sentence_id": 0})
        ))

        self.assertEqual(False, features.not_pronoun_distance(
            Mention(None, Span(100, 100), {"tokens": ["them"], "type": "PRO", "citation_form": "they", "sentence_id": 1}),
            Mention(None, Span(0, 0), {"tokens": ["company"], "type": "NOM", "sentence_id": 0})
        ))

    def test_not_speaker(self):
        self.assertEqual(True, features.not_speaker(
            Mention(None, Span(3, 3), {"tokens": ["I"], "type": "PRO", "citation_form": "i", "speaker": "snafu"}),
            Mention(None, Span(0, 0), {"tokens": ["I"], "type": "PRO", "citation_form": "i", "speaker": "foo"}),
        ))

        self.assertEqual(True, features.not_speaker(
            Mention(None, Span(3, 3), {"tokens": ["us"], "type": "PRO", "citation_form": "we", "speaker": "snafu"}),
            Mention(None, Span(0, 0), {"tokens": ["we"], "type": "PRO", "citation_form": "we", "speaker": "foo"}),
        ))

        self.assertEqual(False, features.not_speaker(
            Mention(None, Span(3, 3), {"tokens": ["you"], "type": "PRO", "citation_form": "you", "speaker": "snafu"}),
            Mention(None, Span(0, 0), {"tokens": ["I"], "type": "PRO", "citation_form": "i", "speaker": "foo"}),
        ))

        self.assertEqual(True, features.not_speaker(
            Mention(None, Span(3, 3), {"tokens": ["you"], "type": "PRO", "citation_form": "you", "speaker": "snafu"}),
            Mention(None, Span(0, 0), {"tokens": ["I"], "type": "PRO", "citation_form": "i", "speaker": "snafu"}),
        ))

    def test_not_embedding(self):
        self.assertEqual(True, features.not_embedding(
            Mention(None, Span(3, 3), {"tokens": ["it"], "type": "PRO", "fine_type": "PERS"}),
            Mention(None, Span(0, 4), {"tokens": ["the", "company", "which", "it", "bought"], "type": "NOM"})
        ))

        self.assertEqual(False, features.not_embedding(
            Mention(None, Span(3, 3), {"tokens": ["its"], "type": "PRO", "fine_type": "POSS_ADJ"}),
            Mention(None, Span(0, 4), {"tokens": ["the", "company", "which", "loves", "its", "success"], "type": "NOM"})
        ))

    def test_not_compatible(self):
        self.assertEqual(True, features.not_compatible(
            Mention(None, Span(0, 0), {"tokens": ["he"], "pos": ["PRP"], "type": "PRO", "number": "SINGULAR", "gender": "MALE", "semantic_class": "PERSON"}),
            Mention(None, Span(1, 1), {"tokens": ["she"], "pos": ["PRP"], "type": "PRO", "number": "SINGULAR", "gender": "FEMALE", "semantic_class": "PERSON"})
        ))

        self.assertEqual(False, features.not_compatible(
            Mention(None, Span(0, 0), {"tokens": ["he"], "pos": ["PRP"], "type": "PRO", "number": "SINGULAR", "gender": "MALE", "semantic_class": "PERSON"}),
            Mention(None, Span(1, 1), {"tokens": ["slawabu"], "pos": ["NN"], "type": "NOM", "number": "UNKNOWN", "gender": "UNKNOWN", "semantic_class": "PERSON"})
        ))

        self.assertEqual(False, features.not_compatible(
            Mention(None, Span(0, 0), {"tokens": ["Jesus"], "pos": ["NNP"], "type": "NAM", "number": "SINGULAR", "gender": "MALE", "semantic_class": "PERSON"}),
            Mention(None, Span(1, 1), {"tokens": ["Jesus"], "pos": ["NNP"], "type": "NAM", "number": "SINGULAR", "gender": "UNKNOWN", "semantic_class": "NORP"})
        ))

    def test_alias(self):
        self.assertEqual(False, features.alias(
            Mention(None, Span(0, 0), {"tokens": ["he"], "type": "PRO", "citation_form": "he"}),
            Mention(None, Span(1, 1), {"tokens": ["he"], "type": "PRO", "citation_form": "he"})
        ))

        self.assertEqual(False, features.alias(
            Mention(None, Span(0, 2), {"head": ["International", "Business", "Machines"], "type": "NAM", "ner": ["ORG", "ORG", "ORG"], "head_index": 2}),
            Mention(None, Span(3, 5), {"head": ["International", "Business", "Machines"], "type": "NAM", "ner": ["ORG", "ORG", "ORG"], "head_index": 2})
        ))

        self.assertEqual(True, features.alias(
            Mention(None, Span(0, 2), {"head": ["International", "Business", "Machines"], "type": "NAM", "ner": ["ORG", "ORG", "ORG"], "head_index": 2}),
            Mention(None, Span(3, 3), {"head": ["IBM"], "type": "NAM", "ner": ["ORG"], "head_index": 0})
        ))

    def test_not_modifier(self):
        self.assertEqual(
            True,
            features.not_modifier(
                Mention(None, Span(10, 16), {"tokens": ["the", "long-awaited", "beginning", "of", "a", "new", "century"], "type": "NOM", "head_span": Span(12, 12), "pos": ["DT", "JJ", "NN", "IN", "DT", "JJ", "NN"]}),
                Mention(None, Span(0, 1), {"tokens": ["the", "beginning"], "type": "NOM", "head_span": Span(1, 1), "pos": ["DT", "NN"]})
            )
        )

        self.assertEqual(
            False,
            features.not_modifier(
                Mention(None, Span(18, 19), {"tokens": ["the", "beginning"], "type": "NOM", "head_span": Span(19, 19), "pos": ["DT", "NN"]}),
                Mention(None, Span(10, 16), {"tokens": ["the", "long-awaited", "beginning", "of", "a", "new", "century"], "type": "NOM", "head_span": Span(12, 12), "pos": ["DT", "JJ", "NN", "IN", "DT", "JJ", "NN"]}),
            )
        )

        self.assertEqual(
            False,
            features.not_modifier(
                Mention(None, Span(18, 19), {"tokens": ["cool", "people"], "type": "NOM", "head_span": Span(19, 19), "pos": ["JJ", "NNS"]}),
                Mention(None, Span(10, 11), {"tokens": ["Cool", "people"], "type": "NOM", "head_span": Span(11, 11), "pos": ["JJ", "NNS"]}),
            )
        )

    def test_get_modifiers(self):
        self.assertEqual(
            set(["long-awaited", "new", "century"]),
            features.get_modifier(Mention(None, Span(10, 16), {"tokens": ["the", "long-awaited", "beginning", "of", "a", "new", "century"], "head_span": Span(12, 12), "pos": ["DT", "JJ", "NN", "IN", "DT", "JJ", "NN"]}))
        )

    def test_get_category_for_alias(self):
        self.assertEqual("PERSON", features.get_category_for_alias("PERSON", "PERSON"))
        self.assertEqual(None, features.get_category_for_alias("LOC", "GPE"))
        self.assertEqual("LOC", features.get_category_for_alias("LOC", "LOC"))
        self.assertEqual(None, features.get_category_for_alias("WORK_OF_ART", "WORK_OF_ART"))

    def test_loc_alias(self):
        self.assertEqual(True, features.loc_alias(["Washington"], ["Washington", ",", "DC"]))
        self.assertEqual(True, features.loc_alias(["NJ"], ["New", "Jersey"]))

    def test_org_alias(self):
        self.assertEqual(True, features.org_alias(["IBM"], ["International", "Business", "Machines"]))
        self.assertEqual(True, features.org_alias(["Lomas", "Corp"], ["Lomas"]))

    def test_person_alias(self):
        self.assertEqual(True, features.person_alias(["Bill", "Clinton"], ["Clinton"]))
        self.assertEqual(True, features.person_alias(["Hillary", "Clinton"], ["Hillary"]))
        self.assertEqual(True, features.person_alias(["Mr.", "Clinton"], ["Bill", "Clinton"]))
        self.assertEqual(False, features.person_alias(["Mr.", "Bush"], ["Bill", "Clinton"]))
        self.assertEqual(False, features.person_alias(["President", "Barack", "Obama"], ["US", "President", "Obama"]))
        self.assertEqual(True, features.person_alias(["President", "Barack", "Obama"], ["President", "Obama"]))
        self.assertEqual(True, features.person_alias(["VOA", "'s", "Barry", "Wood"], ["Barry", "Wood"]))

    def test_starts_with(self):
        self.assertEqual(True, features.starts_with(["Washington"], ["Washington", ",", "DC"]))
        self.assertEqual(True, features.starts_with(["Lomas", "Corp"], ["Lomas"]))
        self.assertEqual(False, features.starts_with(["Lomas", "Corp"], ["Washington"]))

    def test_is_abbreviation(self):
        self.assertEqual(True, features.is_abbreviation(["IBM"], ["I.B.M."]))
        self.assertEqual(True, features.is_abbreviation(["IBM"], ["International", "Business", "Machines"]))
        self.assertEqual(True, features.is_abbreviation(["Darkhorse", "Productions", "Inc."], ["DP"]))
        self.assertEqual(False, features.is_abbreviation(["Darkhorse", "Productions", "Inc."], ["IBM"]))

    def test_get_acronyms(self):
        self.assertEqual(("International Business Machines", "IBM", "I.B.M."),
                         features.get_acronyms(["International", "Business", "Machines"]))

        self.assertEqual(("Darkhorse Productions", "DP", "D.P."),
                         features.get_acronyms(["Darkhorse", "Productions", "Inc."]))

        self.assertEqual(("Darkhorse Productions", "DP", "D.P."),
                         features.get_acronyms(["Darkhorse", "Productions", "Inc"]))



if __name__ == '__main__':
    unittest.main()



""" Represent and manipulate text collections as a list of documents."""

from collections import defaultdict
import multiprocessing

from cort.analysis import data_structures
from cort.core import documents
from cort.core import spans

__author__ = 'smartschat'


def from_string(string):
    return documents.CoNLLDocument(string)


class Corpus:
    """Represents a text collection (a corpus) as a list of documents.

    Such a text collection can also be read from data, and be supplemented with
    antecedent information.

    Attributes:
        description(str): A human-readable description of the corpus.
        documents (list(Document)): A list of CoNLL documents.
    """

    def __init__(self, description, corpus_documents):
        """Construct a Corpus from a description and a list of documents.

        Args:
            description (str): A human-readable description of the corpus.
            documents (list(Document)): A list of documents.
        """
        self.description = description
        self.documents = corpus_documents

    def __iter__(self):
        """Return an iterator over documents in the corpus.

        Returns:
            An iterator over CoNLLDocuments.
        """
        return iter(self.documents)

    @staticmethod
    def from_file(description, coref_file):
        """Construct a new corpus from a description and a file.

        The file must contain documents in the format for the CoNLL shared
        tasks on coreference resolution, see
        http://conll.cemantix.org/2012/data.html.

        Args:
            description (str): A human-readable description of the corpus.
            coref_file (file): A text file of documents in the CoNLL format.

        Returns:
            Corpus: A corpus consisting of the documents described in
            coref_file
        """

        if coref_file is None:
            return []

        document_as_strings = []

        current_document = ""

        for line in coref_file.readlines():
            if line.startswith("#begin") and current_document != "":
                document_as_strings.append(current_document)
                current_document = ""
            current_document += line

        document_as_strings.append(current_document)

        return Corpus(description, sorted([from_string(doc) for doc in
                                           document_as_strings]))



    def write_to_file(self, file):
        """Write a string representation of the corpus to a file,

        Args:
            file (file): The file the corpus should be written to.
        """
        for document in self.documents:
            file.write(document.get_string_representation())

    def write_antecedent_decisions_to_file(self, file):
        """Write antecedent decisions in the corpus to a file.

        For the format, have a look at the documenation for
        read_antecedent_decisions in this class.

        Args:
            file (file): The file the antecedent decisions should be written
                to.
        """
        for document in self.documents:
            document.write_antecedent_decisions_to_file(file)

    def read_antecedents(self, file):
        """Augment corpus with antecedent decisions read from a file.

        The attribute annotated_mentions is overwritten by mentions read in
        from the antecedents file. Input files should have one antecedent
        decision per line, where entries are separated by tabs. The format is

            doc_identifier   (anaphor_start, anaphor_end)    (ante_start, ante_end)

        where
            - doc_id is the identifier in the first line of an CoNLL document
              after #begin document, such as (bc/cctv/00/cctv_0000); part 000
            - anaphor_start is the position in the document where the anaphor
              begins (counting from 0),
            - anaphor_end is the position where the anaphor ends (inclusive),
            - ante_start, ante_end analogously for the antecedent.

        Args:
            file (file): The file the antecedent decisions should be written
                to.
        """
        doc_identifier_to_pairs = defaultdict(list)
        for line in file.readlines():
            splitted = line.split("\t")
            doc_id = splitted[0]
            span_anaphor = splitted[1]
            span_antecedent = splitted[2]
            doc_identifier_to_pairs[doc_id].append(
                (spans.Span.parse(span_anaphor), spans.Span.parse(
                    span_antecedent)))

        for doc in self.documents:
            pairs = sorted(doc_identifier_to_pairs[doc.identifier])
            doc.get_annotated_mentions_from_antecedent_decisions(pairs)

    def read_coref_decisions(self,
                             mention_entity_mapping,
                             antecedent_mapping=None):
        """Augment corpus with coreference and antecedent decisions..

        Set set_id attribute and antecedent information for system mentions.

        Args:
            mention_entity_mapping (dict(Mention, int)): A mapping of mentions
                to entity identifiers.
            antecedent_mapping (dict(Mention, Mention)): A mapping of mentions
                to their antecedent. Optional..
        """
        for doc in self.documents:
            for mention in doc.system_mentions:
                if mention in mention_entity_mapping:
                    mention.attributes["set_id"] = \
                        mention_entity_mapping[mention]
                    if antecedent_mapping and mention in antecedent_mapping:
                        antecedent = antecedent_mapping[mention]
                        mention.attributes['antecedent'] = antecedent
                        mention.document.antecedent_decisions[mention.span] = \
                            antecedent.span

    def get_antecedent_decisions(self, which_mentions="annotated"):
        """ Get all antecedent decisions in this corpus.

        Args:
            which_mentions (str): Either "annotated" or "system". Defaults to
                "system". Signals whether to consider annotated mentions or
                system mentions.

        Returns:
            StructuredCoreferenceAnalysis: A StructuredCoreferenceAnalysis
                containing all antecedent decisions. Can be accessed like a
                dict. If this is assigned a a variable ``x``, the
                decisions can be accessed via ``x[self.description][
                "decisions"]["all"]``, where ``self.description`` is the
                ``description`` attribute of the corpus (e.g. ``x["pair"][
                "decisions"]["all"])..
        """
        antecedent_decisions = {
            self.description: {
                "decisions": {
                    "all": {}
                }
            }
        }

        all_decisions = set()

        for doc in self.documents:
            doc_decisions = doc.get_antecedent_decisions(which_mentions)
            for ana, ante in doc_decisions.items():
                all_decisions.add((ana, ante))

        antecedent_decisions[self.description]["decisions"]["all"] = \
            data_structures.EnhancedSet(all_decisions)

        return data_structures.StructuredCoreferenceAnalysis(
            antecedent_decisions, {self.description: self}, None)

    def are_coreferent(self, m, n):
        """ Compute whether two mentions are coreferent in this corpus.

        One use case of this function is when ``m`` and ``n`` belong to a
        different corpus object, but you are interested in whether they are
        coreferent according to the annotation present in this corpus.

        Args:
            m (Mention): A mention.
            n (Mention): Another mention.

        Returns:
            True if ``m`` and ``n`` are coreferent according to the annotation
            present in this corpus, False otherwise.
        """
        if m.document != n.document:
            return False
        elif m.document not in self.documents:
            return False
        else:
            doc = self.documents[self.documents.index(m.document)]

            if m.span not in doc.spans_to_annotated_mentions or \
               n.span not in doc.spans_to_annotated_mentions:
                return False

            m_in_this_corpus = doc.spans_to_annotated_mentions[m.span]
            n_in_this_corpus = doc.spans_to_annotated_mentions[n.span]

            return m_in_this_corpus.is_coreferent_with(n_in_this_corpus)


import unittest

import nltk

from cort.core import head_finders
from cort.core import spans


__author__ = 'smartschat'


class TestHeadFinder(unittest.TestCase):
    def setUp(self):
        self.head_finder = head_finders.HeadFinder()

    def test_get_head_np(self):
        self.assertEqual(nltk.ParentedTree("NNS", ["police"]), self.head_finder.get_head(nltk.ParentedTree.fromstring("(NP (JJ Local) (NNS police))")))
        self.assertEqual(nltk.ParentedTree("NN", ["shop"]), self.head_finder.get_head(nltk.ParentedTree.fromstring("(NP (JJ Local) (NN shop))")))
        self.assertEqual(nltk.ParentedTree("NNP", ["NBC"]), self.head_finder.get_head(nltk.ParentedTree.fromstring("(NP (NNP NBC) (POS 's))")))
        self.assertEqual(nltk.ParentedTree("NN", ["wedding"]), self.head_finder.get_head(nltk.ParentedTree.fromstring("(NP (NP (NP (PRP$ his) (NN brother) (POS 's)) (NN wedding)) (PP (IN in) (NP (NNP Khan) (NNPS Younes))))")))
        self.assertEqual(nltk.ParentedTree("NNP", ["Taiwan"]), self.head_finder.get_head(nltk.ParentedTree.fromstring("(NP (NNP Taiwan) (POS 's))")))
        self.assertEqual(nltk.ParentedTree("NN", ["port"]), self.head_finder.get_head(nltk.ParentedTree.fromstring("(NP (NP (NP (NNP Yemen) (POS 's)) (NN port)) (PP (IN of) (NP (NNP Aden))))")))

    def test_get_head_vp(self):
        self.assertEqual(nltk.ParentedTree("VB", ["shoot"]), self.head_finder.get_head(nltk.ParentedTree.fromstring("(VP (VB shoot))")))

    def test_get_head_nml(self):
        self.assertEqual(nltk.ParentedTree("NN", ["curtain"]), self.head_finder.get_head(nltk.ParentedTree.fromstring("(NML (NN air) (NN curtain))")))

    def test_get_head_adjp(self):
        self.assertEqual(nltk.ParentedTree("JJ" ,["twelfth"]), self.head_finder.get_head(nltk.ParentedTree.fromstring("(ADJP (JJ twelfth) (CC and) (JJ thirteenth))")))

    def test_get_head_qp(self):
        self.assertEqual(nltk.ParentedTree("CD", ["forty"]), self.head_finder.get_head(nltk.ParentedTree.fromstring("(QP (CD forty) (HYPH -) (CD five))")))

    def test_get_head_whadvp(self):
        self.assertEqual(nltk.ParentedTree("WRB", ["how"]), self.head_finder.get_head(nltk.ParentedTree.fromstring("(WHADVP (WRB how))")))

    def test_get_head_s(self):
        parse = """(S
  (S
    (NP
      (NP
        (DT The)
        (ADJP (RBS most) (JJ important))
        (JJ Taiwanese)
        (JJ musical)
        (NN master))
      (PP (IN of) (NP (DT the) (JJ last) (JJ half) (NN century)))))
  (, ,)
  (NP (PRP he))
  (VP
    (VBD was)
    (NP
      (NP (DT a) (JJ beloved) (NN teacher))
      (PP (IN to) (NP (JJ many)))))
  (. .))"""

        self.assertEqual(nltk.ParentedTree("VBD", ["was"]), self.head_finder.get_head(nltk.ParentedTree.fromstring(parse)))

        parse_2 = "(S (`` `) (NP (NNP Bus) (NNP Stop) (POS ')))"

        self.assertEqual(nltk.ParentedTree("NNP", ["Stop"]),
                         self.head_finder.get_head(
                             nltk.ParentedTree.fromstring(parse_2)))

    def test_get_head_advp(self):
        self.assertEqual(nltk.ParentedTree("RB", ["here"]), self.head_finder.get_head(nltk.ParentedTree.fromstring("(ADVP (RB here))")))

    def test_get_head_whnp(self):
        self.assertEqual(nltk.ParentedTree("WP", ["who"]), self.head_finder.get_head(nltk.ParentedTree.fromstring("(WHNP (WP who))")))

    def test_get_head_sbar(self):
        parse = """(SBAR
  (WHNP (WP who))
  (S
    (VP
      (VBD had)
      (VP
        (VBN had)
        (NP (NP (JJ enough)) (PP (IN of) (NP (NN schooling))))))))"""

        self.assertEqual(nltk.ParentedTree("WP", ["who"]), self.head_finder.get_head(nltk.ParentedTree.fromstring(parse)))

    def test_get_head_pp(self):
        parse = """(PP
  (IN of)
  (NP
    (NP (NNS thousands))
    (PP (IN of) (NP (JJ non-profit) (NNS institutions)))))"""

        self.assertEqual(nltk.ParentedTree("IN", ["of"]), self.head_finder.get_head(nltk.ParentedTree.fromstring(parse)))

    def test_get_head_intj(self):
        self.assertEqual(nltk.ParentedTree("UH", ["oh"]), self.head_finder.get_head(nltk.ParentedTree.fromstring("(INTJ (UH oh) (PRP$ my) (NNP god))")))

    def test_get_head_sq(self):
        self.assertEqual(nltk.ParentedTree("VBP", ["are"]), self.head_finder.get_head(nltk.ParentedTree.fromstring("(SQ (VBP are) (NP (PRP they)) (NP (DT all) (NNS liars)))")))

    def test_get_head_ucp(self):
        self.assertEqual(nltk.ParentedTree("NN", ["trade"]), self.head_finder.get_head(nltk.ParentedTree.fromstring("(UCP (JJ economic) (CC and) (NN trade))")))

    def test_get_head_x(self):
        self.assertEqual(nltk.ParentedTree(":", ["--"]), self.head_finder.get_head(nltk.ParentedTree.fromstring("(X (NNS Men) (CC or) (: --))")))

    def test_get_head_sbarq(self):
        parse = """(SBARQ
  (WHADVP (WRB Where))
  (SQ (MD Should) (NP (NNP Chinese) (NNP Music)) (VP (VB Go)))
  (. ?))"""

        self.assertEqual(nltk.ParentedTree("MD", ["Should"]), self.head_finder.get_head(nltk.ParentedTree.fromstring(parse)))

    def test_get_head_frag(self):
        parse = """(FRAG
  (PP (IN On) (NP (DT the) (NN internet) (NN type)))
  (NP (NNP Iraq))
  (: :)
  (NP (NNP Beyond) (NNP Abu) (NNP Ghraib))
  (: :)
  (NP
    (NP (NN Detention) (CC and) (NN torture))
    (PP (IN in) (NP (NNP Iraq))))
  (. .))"""

        self.assertEqual(nltk.ParentedTree(".", ["."]), self.head_finder.get_head(
            nltk.ParentedTree.fromstring(
            parse)))

    def test_adjust_head_for_nam(self):
        self.assertEqual((spans.Span(0, 1), ["Khan", "Younes"]), head_finders.HeadFinder.adjust_head_for_nam(
            ["Khan", "Younes", "in", "the", "southern", "Ghaza", "Strip"],
            ["NNP", "NNS", "IN", "DT", "JJ", "NNP", "NNP"],
            "GPE"
        ))

        self.assertEqual((spans.Span(0, 1), ["Walter", "Sisulu"]), head_finders.HeadFinder.adjust_head_for_nam(
            ["Walter", "Sisulu"],
            ["NNP", "NNP"],
            "PERSON"
        ))

        self.assertEqual((spans.Span(1, 5), ['vice', 'president', 'Robert', 'W.', 'Reedy']), head_finders.HeadFinder.adjust_head_for_nam(
            ['former', 'vice', 'president', 'Robert', 'W.', 'Reedy'],
            ["JJ", "NN", "NN", "NNP", "NNP", "NNP"],
            "PERSON"
        ))

        self.assertEqual((spans.Span(0, 1), ['Michael', 'Wolf']), head_finders.HeadFinder.adjust_head_for_nam(
            ['Michael', 'Wolf', ',', 'a', 'contributing', 'editor'],
            ["NNP", "NNP", ",", "DT", "VBG", "NN"],
            "PERSON"
        ))

        self.assertEqual((spans.Span(0, 1), ['Mr.', 'Clinton']), head_finders.HeadFinder.adjust_head_for_nam(
            ['Mr.', 'Clinton'],
            ["NNP", "NNP"],
            "NONE"
        ))

        self.assertEqual((spans.Span(0, 0), ['Taiwan']), head_finders.HeadFinder.adjust_head_for_nam(
            ['Taiwan', "'s"],
            ["NNP", "POS"],
            "GPE"
        ))

        self.assertEqual((spans.Span(0, 2), ["Jim", "'s", "Tools"]), head_finders.HeadFinder.adjust_head_for_nam(
            ['Jim', "'s", "Tools"],
            ["NNP", "POS", "NNP"],
            "ORG"
        ))

        self.assertEqual((spans.Span(0, 3), ["Taiwan", "'s", "False", "Cypresses"]), head_finders.HeadFinder.adjust_head_for_nam(
            ["Taiwan", "'s", "False", "Cypresses"],
            ["NNP", "POS", "JJ", "NNP"],
            "NONE"
        ))

    def test_head_rule_cc(self):
        parse = """(NP
        (NP
            (NNS ruin))
        (CC and)
        (NP
            (NNS terror)))
        """

        self.assertEqual(nltk.ParentedTree("CC", ["and"]),
                         self.head_finder.get_head(nltk.ParentedTree.fromstring(
                parse)))

    def test_get_difficult_heads(self):
        parse = """(NP
  (S
    (VP
      (VP
        (VBG recalling)
        (NP (DT the) (JJ Korean) (NN delegation))
        (PP
          (IN to)
          (NP
            (DT the)
            (NNP Korean)
            (NML (NNP Military) (NNP Armistice))
            (NNP Commission))))
      (CC and)
      (VP
        (VBG setting)
        (PRT (RP up))
        (NP
          (NP
            (DT the)
            (NNP Panmunjom)
            (NNP Representative)
            (NNP Office))
          (PP
            (IN of)
            (NP
              (NP (DT the) (NNP Korean) (NNPS People) (POS 's))
              (NNP Army))))
        (PP (IN as) (NP (DT the) (JJ negotiatory) (NN organization))))))
  (, ,)
  (ADVP (FW etc)))"""

        parse2 = """(NP
  (QP (NNS Tens) (IN of) (NNS thousands))
  (PP (IN of) (NP (NNS people))))"""

        parse3 = """(NP
  (PRP he)
  (PRN
    (-LRB- -LRB-)
    (NP
      (NP (DT the) (NN one))
      (SBAR
        (WHNP (WP who))
        (S (VP (VBD tricked) (NP (DT these) (NNS people))))))
    (-RRB- -RRB-)))"""

        parse4 = """(UCP
  (NP (NN %um))
  (CC and)
  (S (NP (PRP you)) (VP (MD can) (VP (ADVP (RB also))))))"""

        self.assertEqual(nltk.ParentedTree("FW", ["etc"]), self.head_finder.get_head(nltk.ParentedTree.fromstring(parse)))
        self.assertEqual(nltk.ParentedTree("NNS", ["Tens"]), self.head_finder.get_head(nltk.ParentedTree.fromstring(parse2)))
        self.assertEqual(nltk.ParentedTree("-LRB-", ["-LRB-"]), self.head_finder.get_head(nltk.ParentedTree.fromstring(parse3)))
        self.assertEqual(nltk.ParentedTree("MD", ["can"]), self.head_finder.get_head(nltk.ParentedTree.fromstring(parse4)))

if __name__ == '__main__':
    unittest.main()



""" Algorithms for computing spanning trees of entity graphs. """


__author__ = 'smartschat'


def precision_system_output(entity, partitioned_entity):
    """ Compute a spanning tree from antecedent information.

    All edges in the spanning tree correspond to anaphor-antecedent pairs. In
    order to access this antecedent information, the attribute "antecedent" of
    the mentions in the entity must be set.

    Args:
        entity (EntityGraph): The EntityGraph for the entity for which the
            spanning tree should be computed.
        partitioned_entity (EntityGraph): A partition of the entity -- not
            used for this algorithm.

    Returns:
        list(Mention, Mention): A list of mention pairs, which constitute the
        edges of the spanning tree. For a pair (m, n), n appears later in
        the text than m.
    """
    edges = []
    for mention in entity.edges:
        # just look at system output
        if ("antecedent" in mention.attributes
                and mention.attributes["antecedent"] in entity.edges[mention]):
            edges.append((mention, mention.attributes["antecedent"]))

    return sorted(edges)


def recall_closest(entity, partitioned_entity):
    """ Compute a spanning tree by always taking the closest mention in the same
    entity.

    Args:
        entity (EntityGraph): The EntityGraph for the entity for which the
            spanning tree should be computed.
        partitioned_entity (EntityGraph): A partition of the entity -- not
            used for this algorithm.

    Returns:
        list(Mention, Mention): A list of mention pairs, which constitute the
        edges of the spanning tree. For a pair (m, n), n appears later in
        the text than m.
    """
    edges = []
    for mention in entity.edges:
        # always take closest (except for first mention in entity, which does
        # not have any antecedent)
        if entity.edges[mention]:
            if mention in partitioned_entity.edges:
                antecedent = sorted(partitioned_entity.edges[mention],
                                    reverse=True)[0]
            else:
                antecedent = sorted(entity.edges[mention], reverse=True)[0]
            edges.append((mention, antecedent))

    return sorted(edges)


def recall_accessibility(entity, partitioned_entity):
    """ Compute a spanning tree by choosing edges according to the accessibility
    of the antecedent.

    First, if a mention has an out-degree of at least one in the partitioned
    entity, take the edge with the closest mention distance as an edge for
    the spanning tree. Otherwise, proceed as follows.

    If a mention m is a proper name or a common noun, choose an antecedent as
    follows:

        - if a proper name antecedent exists, take the closest and output this
          pair as an edge
        - else if a common noun antecedent exists, take the closest and output
          this pair as an edge
        - else take the closest preceding mention and output this pair as an
          edge

    For all other mentions, take the closest preceding mention and output
    this pair as an edge.

    Args:
        entity (EntityGraph): The EntityGraph for the entity for which the
            spanning tree should be computed.
        partitioned_entity (EntityGraph): A partition of the entity -- not
            used for this algorithm.

    Returns:
        list(Mention, Mention): A list of mention pairs, which constitute the
        edges of the spanning tree. For a pair (m, n), n appears later in
        the text than m.
    """
    edges = []
    for mention in entity.edges:
        if entity.edges[mention]:
            # mention is not the first in subentity? take closest!
            if mention in partitioned_entity.edges:
                antecedent = sorted(partitioned_entity.edges[mention],
                                    reverse=True)[0]
            else:
                antecedent = __get_antecedent_by_type(mention,
                                                      entity.edges[mention])

            edges.append((mention, antecedent))

    return sorted(edges)


def __get_antecedent_by_type(mention, candidates):
    # make sure...
    candidates_reversed = sorted(candidates, reverse=True)
    # mention is (demonstrative) pronoun? take closest!
    if (mention.attributes["type"] == "PRO" or
            mention.attributes["type"] == "DEM"):
        return candidates_reversed[0]
    # otherwise chose by type, back off to closest
    elif __get_by_pos(candidates_reversed, "NAM"):
        return __get_by_pos(candidates_reversed, "NAM")
    elif __get_by_pos(candidates_reversed, "NOM"):
        return __get_by_pos(candidates_reversed, "NOM")
    else:
        return candidates_reversed[0]


def __get_by_pos(candidates, pos):
    for mention in candidates:
        if mention.attributes["type"] == pos:
            return mention



__author__ = 'smartschat'


class MultigraphDecoder:
    def __init__(self, multigraph_creator):
        self.coref_multigraph_creator = multigraph_creator

    def decode(self, corpus):
        for doc in corpus:
            for mention in doc.system_mentions:
                mention.attributes["set_id"] = None

            # discard dummy mention
            self.decode_for_one_document(doc.system_mentions[1:])

    def decode_for_one_document(self, mentions):
        multigraph = \
            self.coref_multigraph_creator.construct_graph_from_mentions(
                mentions)

        for mention in mentions:
            antecedent = self.compute_antecedent(mention, multigraph)

            if antecedent is not None:
                if antecedent.attributes["set_id"] is None:
                    antecedent.attributes["set_id"] = \
                        mentions.index(antecedent)

                mention.attributes["set_id"] = antecedent.attributes["set_id"]
                mention.document.antecedent_decisions[mention.span] = \
                    antecedent.span

    @staticmethod
    def compute_antecedent(mention, multigraph):
        weights = []
        for antecedent in multigraph.edges[mention]:
            if not multigraph.edges[mention][antecedent]["negative_relations"]:
                weights.append(
                    (multigraph.get_weight(mention, antecedent), antecedent))

        # get antecedent with highest positive weight, break ties by distance
        if len(weights) > 0 and sorted(weights)[-1][0] > 0:
            return sorted(weights)[-1][1]



from cort.core.util import clean_via_pos

__author__ = 'smartschat'

import unittest


class TestUtil(unittest.TestCase):
    def test_clean_via_pos(self):
        self.assertEqual(
            ["newly-elect", "leader", "wife"],
            clean_via_pos(
                ["the", "newly-elect", "leader", "'s", "wife"],
                ["DT", "JJ", "NN", "POS", "NN"]))


if __name__ == '__main__':
    unittest.main()



""" Manage mentions and their attributes. """

from cort.core import mention_property_computer
from cort.core import spans


__author__ = 'smartschat'


class Mention:
    """ A mention is an expression in a document which is potentially referring.

    Attributes:
        document (CoNLLDocument): The document the mention belongs to.
        span (Span): The span of the mention in its document. If for example
            the span is (3, 4), then the mention starts at the 3rd token in
            the document and ends at the 4th (inclusive).
        attributes (dict(str, object)): A mapping of attribute names to
            attribute values. When creating a document from a text, The
            following attributes are used:

                - tokens (list(str)): the tokens of the mention,
                - head (list(str)): the head words of the mention,
                - pos (list(str)): the part-of-speech tags of the mention,
                - ner (list(str)): the named entity tags of the mention,
                  as found in the data,
                - type (str): the mention type, one of

                    - NAM (proper name),
                    - NOM (common noun),
                    - PRO (pronoun),
                    - DEM (demonstrative pronoun),
                    - VRB (verb),

                - fine_type (str): only set when the mention is a nominal or a
                  pronoun, for nominals values

                    - DEF (definite noun phrase) or
                    - INDEF,

                  for pronouns values

                    - PERS_NOM (personal pronoun, nominative case),
                    - PERS_ACC (personal pronoun, accusative),
                    - REFL (reflexive pronoun),
                    - POSS (possessive pronoun) or
                    - POSS_ADJ (possessive adjective, e.g. 'his'),

                - citation_form (str): only set if the mention is a pronoun,
                  then the canonical form of the pronoun, i.e. one of
                  i, you, he, she, it, we, they,
                - grammatical_function (str): either SUBJECT, OBJECT or OTHER,
                - number (str): either SINGULAR, PLURAL or UNKNOWN,
                - gender (str): either MALE, FEMALE, NEUTRAL, PLURAL or UNKNOWN,
                - semantic_class (str): either PERSON, OBJECT or UNKNOWN,
                - sentence_id (int): the sentence id of the mention's sentence
                  (starting at 0),
                - parse_tree (nltk.ParentedTree): the parse tree of the mention,
                - speaker (str): the speaker of the mention,
                - antecedent (Mention): the antecedent of the mention
                  (intially None),
                - annotated_set_id (str): the set id of the mention as found
                  in the data,
                - set_id (str): the set id of the mention computed by a
                  coreference resolution approach (initially None),
                - head_span (Span): the span of the head (in the document),
                - head_index (int): the mention-internal index of the start of
                  the head,
                - is_apposition (bool): whether the mention contains an
                  apposition,
                - head_as_lowercase_string (str): the head lowercased and as a
                  string,
                - tokens_as_lowercase_string (str): all tokens of the mention
                  lowercased and as a string,
                - first_in_gold_entity (bool): whether the mention is the first
                  mention in its gold entity (for system mentions, this is
                  also true if no preceding mention in the same entity was
                  found by the mention extractor),

    """
    def __init__(self, document, span, attributes):
        """ Initialize a mention in a document.

        Args:
            document (CoNLLDocument): The document the mention belongs to.
            span (Span): The span of the mention in its document.
            attributes (dict(str, object)): A mapping of attribute names to
                attribute values (see the class documentation for more
                information).
        """
        self.document = document
        self.span = span
        self.attributes = attributes

    @staticmethod
    def dummy_from_document(document):

        return Mention(document, None, {
            "is_dummy": True,
            "annotated_set_id": None,
            "tokens": [],
            "first_in_gold_entity": True
        })

    def is_dummy(self):
        return "is_dummy" in self.attributes and self.attributes["is_dummy"]

    @staticmethod
    def from_document(span, document, first_in_gold_entity=False):
        """
        Create a mention from a span in a document.

        All attributes of the mention are computed from the linguistic
        information found in the document. For information about the
        attributes, see the class documentation.

        Args:
            document (CoNLLDocument): The document the mention belongs to.
            span (Span): The span of the mention in the document.

        Returns:
            Mention: A mention extracted from the input span in the input
            document.
        """

        i, sentence_span = document.get_sentence_id_and_span(span)

        attributes = {
            "tokens": document.tokens[span.begin:span.end + 1],
            "pos": document.pos[span.begin:span.end + 1],
            "ner": document.ner[span.begin:span.end + 1],
            "sentence_id": i,
            "parse_tree": mention_property_computer.get_relevant_subtree(
                span, document),
            "speaker": document.speakers[span.begin],
            "antecedent": None,
            "set_id": None,
            "first_in_gold_entity": first_in_gold_entity
        }

        if span in document.coref:
            attributes["annotated_set_id"] = document.coref[span]
        else:
            attributes["annotated_set_id"] = None

        attributes["is_apposition"] = \
            mention_property_computer.is_apposition(attributes)

        attributes["grammatical_function"] = \
            mention_property_computer.get_grammatical_function(attributes)

        (head, in_mention_span, head_index) = \
            mention_property_computer.compute_head_information(attributes)

        attributes["head"] = head
        attributes["head_span"] = spans.Span(
            span.begin + in_mention_span.begin,
            span.begin + in_mention_span.end
        )
        attributes["head_index"] = head_index

        attributes["type"] = mention_property_computer.get_type(attributes)
        attributes["fine_type"] = mention_property_computer.get_fine_type(
            attributes)

        if attributes["type"] == "PRO":
            attributes["citation_form"] = \
                mention_property_computer.get_citation_form(
                    attributes)

        attributes["number"] = \
            mention_property_computer.compute_number(attributes)
        attributes["gender"] = \
            mention_property_computer.compute_gender(attributes)

        attributes["semantic_class"] = \
            mention_property_computer.compute_semantic_class(attributes)

        attributes["head_as_lowercase_string"] = " ".join(attributes[
            "head"]).lower()

        attributes["tokens_as_lowercase_string"] = " ".join(attributes[
            "tokens"]).lower()

        dep_tree = document.dep[i]

        index = span.begin + head_index - sentence_span.begin

        governor_id = dep_tree[index].head - 1

        if governor_id == -1:
            attributes["governor"] = "NONE"
        else:
            attributes["governor"] = dep_tree[governor_id].form.lower()

        attributes["ancestry"] = Mention._get_ancestry(dep_tree, index)

        attributes["deprel"] = dep_tree[index].deprel

        return Mention(document, span, attributes)

    @staticmethod
    def _get_ancestry(dep_tree, index, level=0):
        if level >= 2:
            return ""
        else:
            governor_id = dep_tree[index].head - 1

            direction = "L"

            if governor_id > index:
                direction = "R"

            if governor_id == -1:
                return  "-" + direction + "-NONE"
            else:
                return "-" + direction + "-" + dep_tree[governor_id].pos + \
                    Mention._get_ancestry(dep_tree, governor_id, level+1)



    def __lt__(self, other):
        """ Check whether this mention is less than another mention.

        ``self < other`` if and only if ``self.span < other.span``, that is,

            - this mention has span None (is a dummy mention), and the other
              mention has a span which is not None, or
            - this mentions begins before the other mention, or
            - the mentions begin at the same position, but this mention ends
              before the other mention.

        Args:
            other (Mention): A mention.

        Returns:
            True if this mention is less than other, False otherwise.
        """

        if self.span is None:
            return other.span is not None
        elif other.span is None:
            return False
        else:
            return self.span < other.span

    def __eq__(self, other):
        """ Check for equality.

        Two mentions are equal if they are in the same document and have the
        same span.

        Args:
            other (Mention): A mention.

        Returns:
            True if the mentions are in the same document and have the same
            span.
        """
        if isinstance(other, self.__class__):
            return self.span == other.span and self.document == other.document
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        if self.document is None:
            return hash((self.span.begin, self.span.end))
        elif self.span is None:
            return hash(self.document.identifier)
        else:
            return hash((self.document.identifier,
                         self.span.begin,
                         self.span.end))

    def __str__(self):
        return (repr(self.document) +
                ", " +
                str(self.span) +
                ": "
                + " ".join(self.attributes["tokens"]))

    def __repr__(self):
        return (repr(self.document) +
                ", " +
                str(self.span) +
                ": " +
                str(self.attributes["tokens"]))

    def get_context(self, window):
        """ Get the context in a window around the mention.

        Args:
            window (int): An integer specifying the size of the window.

        Returns:
            list(str): The tokens in a window of around the mention.

            In particular, get ``window`` tokens to the right or left of the
            mention,, depending on the sign of ``window``: if the sign is +,
            then to the right, if the sign is -, then to the left. Return
            None if the window is not contained in the document.
        """
        if window < 0 <= window + self.span.begin:
            return self.document.tokens[
                self.span.begin + window:self.span.begin]
        elif (window > 0 and self.span.end + window + 1
                <= len(self.document.tokens)):
            return self.document.tokens[
                self.span.end + 1:self.span.end + window + 1]

    def is_coreferent_with(self, m):
        """ Return whether this mention is coreferent with another mention.

        Args:
            m (Mention): Another mention.

        Returns:
            True if m and this mention are coreferent (are in the same document
            and have the same annotated set id), False otherwise.
        """

        self_set_id = self.attributes['annotated_set_id']
        m_set_id = m.attributes['annotated_set_id']

        if self.document is None and m.document is None:
            return self_set_id is not None and self_set_id == m_set_id
        elif self.is_dummy():
            return m.is_dummy()
        elif m.is_dummy():
            return self.is_dummy()
        else:
            return self.document == m.document \
                and self_set_id is not None \
                and self_set_id == m_set_id

    def decision_is_consistent(self, m):
        """ Return whether the decision to put this mention and m into the
        same entity is consistent with the gold annotation.

        The decision is consistent if one of the following conditions holds:

            - the mentions are coreferent,
            - one of the mentions is the dummy mention, and the other mention
              does not have a preceding mention that it is coreferent with.

        Args:
            m (Mention): Another mention.

        Returns:
            True if m and this mention are consistent according to the
            definition above, False otherwise.
        """

        if self.is_coreferent_with(m):
            return True
        elif self.is_dummy():
            return m.attributes['annotated_set_id'] is None \
                   or m.attributes["first_in_gold_entity"]
        elif m.is_dummy():
            return self.attributes['annotated_set_id'] is None \
                   or self.attributes["first_in_gold_entity"]
        else:
            return False



""" Plot error analysis statistics. """

from __future__ import division


from matplotlib import pyplot
from matplotlib import cm

import numpy

from pylab import rcParams


__author__ = 'martscsn'


def plot(data,
         title,
         xlabel,
         ylabel,
         filename=None):
    """ Plot error analysis statistics.

    In particular, plot a bar chart for the numbers described in ``data``.

    Args:
        data (list(str, list((str,int)))): The data to be plotted. The ith entry
            of this list contains the name which will appear in the legend,
            and a list of (category, count) pairs. These are the individual
            data points which will be plotted.
        title (str): Title of the plot.
        xlabel (str): Label of the x axis.
        ylabel (str): Label of the y axis.
        filename (str, optional): If set, write plot to ``filename``.

    Example::
        pair_errs = errors["pair"]["recall_errors"]["all"]
        tree_errs = errors["tree"]["recall_errors"]["all"]

        plot(
            [("pair", [(cat, len(pair_errs[cat])) for cat in pair_errs.keys()]),
            ("tree", [(cat, len(tree_errs[cat])) for cat in tree_errs.keys()])],
            "Recall Errors",
            "Type of anaphor",
            "Number of Errors")
    """

    rcParams['xtick.major.pad'] = '12'
    rcParams['ytick.major.pad'] = '12'

    fig, ax = pyplot.subplots()

    systems = []
    categories = []

    colors = cm.Accent(numpy.linspace(0, 1, len(data)))

    bars_for_legend = []

    for i, system_data in enumerate(data):
        system_name, categories_and_numbers = system_data
        systems.append(system_name)

        for j, cat_and_number in enumerate(categories_and_numbers):
            category, number = cat_and_number

            if category not in categories:
                categories.append(category)

            bar = ax.bar(2*j + i*(1/len(data)), number, color=colors[i],
                         width=1/len(data), label=system_name)

            if j == 0:
                bars_for_legend.append(bar)

    xticks = [2*k + 0.5 for k in range(0, len(categories))]

    pyplot.title(title, fontsize=28)
    pyplot.xlabel(xlabel, fontsize=24)
    pyplot.ylabel(ylabel, fontsize=24)

    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)

    ax.get_xaxis().tick_bottom()
    ax.get_yaxis().tick_left()

    ax.set_xticklabels(categories)
    ax.set_xticks(xticks)

    pyplot.tick_params(axis='both', which='major', labelsize=20)

    if filename:
        legend = ax.legend(bars_for_legend, systems,
                           loc='upper right', bbox_to_anchor=(1.2, 1.2))

        fig.savefig(filename, bbox_extra_artists=(legend,), bbox_inches='tight')
    else:
        legend = ax.legend(bars_for_legend, systems, loc='upper right')
        legend.draggable()

        fig.show()



""" Contains features for coreference resolution.

Features takes as input either a mention or a pair of mentions, and always
output a tuple ``(feature_name, value)``. ``feature_name`` is the name of the
feature (a string), and ``value`` is the value of the feature. ``value`` can be
either a string, a bool, or a scalar.
"""

from __future__ import division


import re


from cort.core import spans


__author__ = 'smartschat'


def fine_type(mention):
    """ Compute fine-grained type of a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        The tuple ('fine_type', TYPE), where where TYPE is one of 'NAM', 'DEF',
        'INDEF', 'DEM', 'VRB', 'i', 'you', 'he', 'she', 'it', 'we', 'they'
        and 'NONE'.
    """
    if mention.attributes["type"] == "NOM":
        mention_fine_type = mention.attributes["fine_type"]
    elif mention.attributes["type"] == "PRO":
        mention_fine_type = mention.attributes["citation_form"]
    else:
        mention_fine_type = mention.attributes["type"]

    if not mention_fine_type:
        mention_fine_type = "NONE"

    return "fine_type", mention_fine_type


def gender(mention):
    """ Compute gender of a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        The tuple ('gender', GENDER), where GENDER is one of 'MALE',
        'FEMALE', 'NEUTRAL', 'PLURAL' and 'UNKNOWN'.
    """
    return "gender", mention.attributes["gender"]


def number(mention):
    """ Compute number of a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        The tuple ('number', NUMBER), where NUMBER is one of 'SINGULAR',
        'PLURAL' and 'UNKNOWN'.
    """
    return "number", mention.attributes["number"]


def sem_class(mention):
    """ Compute semantic class of a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        The tuple ('sem_class', SEM_CLASS), where SEM_CLASS is one of
        'PERSON', 'OBJECT', 'NUMERIC' and 'UNKNOWN'.
    """
    return "sem_class", mention.attributes["semantic_class"]


def gr_func(mention):
    """ Compute grammatical function of a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        The tuple ('gr_func', GR_FUNC), where GR_FUNC is one of 'SUBJECT',
        'OBJECT' and 'OTHER'.
    """
    return "gr_func", mention.attributes["grammatical_function"]


def governor(mention):
    """ Compute governor of a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        The tuple ('governor', GOVERNOR), where GOVERNOR is the governor
        of the mention.
    """
    return "governor", mention.attributes["governor"].lower()


def deprel(mention):
    """ Compute dependency relation of a mention to its governor.

    Args:
        mention (Mention): A mention.

    Returns:
        The tuple ('deprel', DEPREL), where DEPREL is the dependency relation
        of the mention to its governor.
    """
    return "deprel", mention.attributes["deprel"]


def head(mention):
    """ Compute head of a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        The tuple ('head', HEAD), where HEAD is the (lowercased) head of
        the mention.
    """
    return "head", mention.attributes["head_as_lowercase_string"]


def head_ner(mention):
    """ Compute named entity tag of a mention's head.

    Args:
        mention (Mention): A mention.

    Returns:
        The tuple ('ner', NER), where NER is the named entity tag of the
        mention's head word. If the mention is not a named entity, NER is
        set to 'NONE'.
    """
    return "ner", mention.attributes["ner"][mention.attributes["head_index"]]


def length(mention):
    """ Compute length of a mention in tokens.

    Args:
        mention (Mention): A mention.

    Returns:
        The tuple ('length' LENGTH), where LENGTH is the length of the
        mention in tokens. The length is stored as a string.
    """
    return "length", str(len(mention.attributes["tokens"]))


def first(mention):
    """ Compute the first token of a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        The tuple ('first', TOKEN), where TOKEN is the (lowercased) first
        token of the mention.
    """
    return "first", mention.attributes["tokens"][0].lower()


def last(mention):
    """ Compute the last token of a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        The tuple ('last', TOKEN), where TOKEN is the (lowercased) last token
        of the mention.
    """
    return "last", mention.attributes["tokens"][-1].lower()


def preceding_token(mention):
    """ Compute the token preceding a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        The tuple ('preceding', TOKEN), where TOKEN is the (lowercased) token
        preceding the mention. If no such token exists, set TOKEN to 'NONE'.
    """
    prec = mention.get_context(-1)

    if prec:
        return "preceding", prec[0].lower()
    else:
        return "preceding", "NONE"


def next_token(mention):
    """ Compute the token following a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        The tuple ('next', TOKEN), where TOKEN is the (lowercased) token
        following the mention. If no such token exists, set TOKEN to 'NONE'.
    """
    next_t = mention.get_context(1)

    if next_t:
        return "next", next_t[0].lower()
    else:
        return "next", "NONE"


def ancestry(mention):
    """ Compute the ancestry of a mention.

    We follow the definition of the ancestry by Durrett and Klein (2013). For
    more information, have a look at their paper:

    Greg Durrett and Dan Klein. Easy Victories and Uphill Battles in
    Coreference Resolution. In Proceedings of EMNLP 2013.
    http://anthology.aclweb.org/D/D13/D13-1203.pdf

    Args:
        mention (Mention): A mention.

    Returns:
        The tuple ('ancestry', ANCESTRY), where ANCESTRY is the ancestry of
        the mention.
    """
    return "ancestry", mention.attributes["ancestry"]


def exact_match(anaphor, antecedent):
    """ Compute whether the tokens of two mentions match exactly.

    Args:
        anaphor (Mention): A mention.
        antecedent (Mention): Another mention, preceding the anaphor.

    Returns:
        The tuple ('exact_match', MATCH), where MATCH is True if the mentions
        match exactly (ignoring case), and False otherwise.
    """

    match = anaphor.attributes["tokens_as_lowercase_string"] == \
            antecedent.attributes["tokens_as_lowercase_string"]

    return "exact_match", match


def head_match(anaphor, antecedent):
    """ Compute whether the heads of two mentions match.

    Args:
        anaphor (Mention): A mention.
        antecedent (Mention): Another mention, preceding the anaphor.

    Returns:
        The tuple ('head_match', MATCH), where MATCH is True if the heads of
        the mentions match exactly (ignoring case), and False otherwise.
    """
    match = anaphor.attributes["head_as_lowercase_string"] == \
            antecedent.attributes["head_as_lowercase_string"]

    return "head_match", match


def tokens_contained(anaphor, antecedent):
    """ Compute whether one mention is a substring of the other mention.

    Args:
        anaphor (Mention): A mention.
        antecedent (Mention): Another mention, preceding the anaphor.

    Returns:
        The tuple ('tokens_contained', CONTAINED), where CONTAINED is True if
        the anaphor contains the antecedent, or vice versa (ignoring case).
    """
    ana_tokens = anaphor.attributes["tokens_as_lowercase_string"]
    ante_tokens = antecedent.attributes["tokens_as_lowercase_string"]

    contained = ana_tokens in ante_tokens or ante_tokens in ana_tokens

    return "tokens_contained", contained


def head_contained(anaphor, antecedent):
    """ Compute whether one mention's head is a substring of the other
    mention's head.

    Args:
        anaphor (Mention): A mention.
        antecedent (Mention): Another mention, preceding the anaphor.

    Returns:
        The tuple ('head_contained', CONTAINED), where CONTAINED is True if
        the anaphor's head contains the antecedent's head, or vice versa
        (ignoring case).
    """
    ana_head = anaphor.attributes["head_as_lowercase_string"]
    ante_head = antecedent.attributes["head_as_lowercase_string"]

    contained = ana_head in ante_head or ante_head in ana_head

    return "head_contained", contained


def sentence_distance(anaphor, antecedent):
    """ Compute the sentence distance between two mentions (capped at 5).

    Args:
        anaphor (Mention): A mention.
        antecedent (Mention): Another mention, preceding the anaphor.

    Returns:
        The tuple ('sentence_distance', DIST), where DIST is one of '0', '1',
        '2', '3', '4' and '>=5'.
    """
    return "sentence_distance", __compute_sentence_distance(anaphor,
                                                            antecedent)


def token_distance(anaphor, antecedent):
    """ Compute the token distance between two mentions (capped at 10).

    Args:
        anaphor (Mention): A mention.
        antecedent (Mention): Another mention, preceding the anaphor.

    Returns:
        The tuple ('token_distance'=DIST), where DIST is one of '0', '1',
        '2', '3', '4' and '>=10'.
    """
    return "token_distance", __compute_token_distance(anaphor, antecedent)

def alias(anaphor, antecedent):
    """ Compute whether the mentions are aliases of each other.

    Args:
        anaphor (Mention): A mention.
        antecedent (Mention): Another mention, preceding the anaphor.

    Returns:
        The tuple ('alias', ALIAS), where ALIAS is True if anaphor and
        antecedent are in an alias relation, False otherwise.
    """
    return "alias", __are_alias(anaphor, antecedent)


def same_speaker(anaphor, antecedent):
    """ Compute whether the speakers  of two mentions are the same.

    Args:
        anaphor (Mention): A mention.
        antecedent (Mention): Another mention, preceding the anaphor.

    Returns:
        The tuple ('same_speaker', SAME), where SAME is True if anaphor and
        antecedent have the same speaker, False otherwise.
    """
    same = anaphor.attributes["speaker"] == antecedent.attributes["speaker"]

    return "same_speaker", same


def embedding(anaphor, antecedent):
    """ Compute whether one mention embeds the other.

    Args:
        anaphor (Mention): A mention.
        antecedent (Mention): Another mention, preceding the anaphor.

    Returns:
        The tuple ('embedding', EMB), where EMB is True if one of the mentions
        embeds the other, False otherwise.
    """
    emb = anaphor.span.embeds(antecedent.span) or \
          antecedent.span.embeds(anaphor.span)

    return "embedding", emb


def modifier(anaphor, antecedent):
    """ Compute modifier agreement.

    In particular, compute whether the anaphor has modifiers that do not appear
    in the antecedent (ignoring demonstratives and prepositions).

    Args:
        anaphor (Mention): A mention.
        antecedent (Mention): Another mention, preceding the anaphor.

    Returns:
        The tuple ('modifier', MOD), where MOD is True if the anaphor has
        modifiers that do not appear in the antecedent, False otherwise.
    """
    mod = not __get_modifier(anaphor).issubset(__get_modifier(antecedent))

    return "modifier", mod


def relative_overlap(anaphor, antecedent):
    """ Compute relative overlap of the mentions (ignoring case).

    For example, "the new president" and "the president" have relative overlap
    2/3.

    Args:
        anaphor (Mention): A mention.
        antecedent (Mention): Another mention, preceding the anaphor.

    Returns:
        The tuple ('relative_overlap', OVERLAP), where OVERLAP (a float) is
        the relative overlap of anaphor and antecedent.
    """
    ana_tokens = set([tok.lower() for tok in anaphor.attributes["tokens"]])
    ante_tokens = set([tok.lower() for tok
                             in antecedent.attributes["tokens"]])

    overlap = len(ana_tokens & ante_tokens)/max(len(ana_tokens),
                                                len(ante_tokens))

    return "relative_overlap", overlap


def __compute_sentence_distance(anaphor, antecedent):
    dist = anaphor.attributes['sentence_id'] - antecedent.attributes[
        'sentence_id']

    if dist >= 5:
        return ">=5"
    else:
        return str(dist)


def __compute_token_distance(anaphor, antecedent):
    dist = anaphor.span.begin - antecedent.span.end
    if dist >= 10:
        return ">=10"
    else:
        return str(dist)


def __get_modifier(mention):
    head_span_in_mention = spans.Span(
        mention.attributes["head_span"].begin - mention.span.begin,
        mention.attributes["head_span"].end - mention.span.begin)

    modifiers = set()

    for index, (token, pos) in enumerate(
            zip(mention.attributes["tokens"], mention.attributes["pos"])):
        if (token.lower() not in ["the", "this", "that", "those", "these",
                                  "a", "an"]
            and pos not in ["POS", "IN"]
            and (index < head_span_in_mention.begin
                 or index > head_span_in_mention.end)):
            modifiers.add(token.lower())

    return modifiers


def __are_alias(anaphor, antecedent):
    if anaphor.attributes["type"] != "NAM" or antecedent.attributes["type"] \
            != "NAM":
        return False
    elif anaphor.attributes["head_as_lowercase_string"] == \
            antecedent.attributes["head_as_lowercase_string"]:
        return False
    else:
        anaphor_cleaned_tokens = anaphor.attributes["head"]
        antecedent_cleaned_tokens = antecedent.attributes["head"]

        category = __get_category_for_alias(
            anaphor.attributes["ner"][
                anaphor.attributes["head_index"]
            ], antecedent.attributes["ner"][
                antecedent.attributes["head_index"]]
        )

        if category == "PERSON":
            return __person_alias(anaphor_cleaned_tokens,
                                  antecedent_cleaned_tokens)
        elif category == "LOC":
            return __loc_alias(anaphor_cleaned_tokens,
                               antecedent_cleaned_tokens)
        elif category == "ORG":
            return __org_alias(anaphor_cleaned_tokens,
                               antecedent_cleaned_tokens)
        else:
            return False


def __get_category_for_alias(anaphor_ner, antecedent_ner):
    if anaphor_ner == "PERSON" and antecedent_ner == "PERSON":
        return "PERSON"
    elif re.match(r"LOC", anaphor_ner) and re.match(r"LOC", antecedent_ner):
        return "LOC"
    elif re.match(r"ORG", anaphor_ner) and re.match(r"(ORG)", antecedent_ner):
        return "ORG"


def __loc_alias(anaphor_cleaned_tokens, antecedent_cleaned_tokens):
    return __starts_with(anaphor_cleaned_tokens, antecedent_cleaned_tokens) or \
        __is_abbreviation(anaphor_cleaned_tokens, antecedent_cleaned_tokens)


def __org_alias(anaphor_cleaned_tokens, antecedent_cleaned_tokens):
    return __starts_with(anaphor_cleaned_tokens, antecedent_cleaned_tokens) or \
        __is_abbreviation(anaphor_cleaned_tokens, antecedent_cleaned_tokens)


def __person_alias(anaphor_cleaned_tokens, antecedent_cleaned_tokens):
    if len(anaphor_cleaned_tokens) == 1 or len(antecedent_cleaned_tokens) == 1:
        return anaphor_cleaned_tokens[0] == antecedent_cleaned_tokens[0] \
            or anaphor_cleaned_tokens[-1] == antecedent_cleaned_tokens[-1]
    elif (len(anaphor_cleaned_tokens) == 2 and anaphor_cleaned_tokens[
        0].lower() in ["mr", "ms", "mr.", "ms."]
        or len(antecedent_cleaned_tokens) == 2 and antecedent_cleaned_tokens[
            0].lower() in ["mr", "ms", "mr.", "ms."]):
        return anaphor_cleaned_tokens[-1] == antecedent_cleaned_tokens[-1]
    elif anaphor_cleaned_tokens[0] == antecedent_cleaned_tokens[0] and \
            anaphor_cleaned_tokens[-1] == antecedent_cleaned_tokens[-1]:
        return True
    elif len(anaphor_cleaned_tokens) > 1 and len(antecedent_cleaned_tokens) > 1:
        return anaphor_cleaned_tokens[-1] == antecedent_cleaned_tokens[-1] and \
            anaphor_cleaned_tokens[-2] == antecedent_cleaned_tokens[-2]

    return False


def __starts_with(anaphor_cleaned_tokens, antecedent_cleaned_tokens):
    for ana_token, ante_token in zip(anaphor_cleaned_tokens,
                                     antecedent_cleaned_tokens):
        if ana_token != ante_token:
            return False

    return True


def __is_abbreviation(anaphor_cleaned_tokens, antecedent_cleaned_tokens):
    if " ".join(anaphor_cleaned_tokens).replace(".", "") == " ".join(
            antecedent_cleaned_tokens).replace(".", ""):
        return True
    else:
        if len(anaphor_cleaned_tokens) > len(antecedent_cleaned_tokens):
            return " ".join(antecedent_cleaned_tokens) in set(
                __get_acronyms(anaphor_cleaned_tokens))
        else:
            return " ".join(anaphor_cleaned_tokens) in set(
                __get_acronyms(antecedent_cleaned_tokens))


def __get_acronyms(cleaned_tokens):
    company_designator = r'assoc|bros|co|coop|corp|devel|inc|llc|ltd\.?'
    tokens_without_designator = [token for token in cleaned_tokens if
                                 not re.match(company_designator,
                                              token.lower())]

    return " ".join(tokens_without_designator), \
           "".join([token[0] for token in tokens_without_designator if
                    token[0].isupper()]), \
           ".".join([token[0] for token in tokens_without_designator if
                     token[0].isupper()]) + "."



import os
import unittest

from cort.core.corpora import Corpus


__author__ = 'smartschat'


class TestCorpora(unittest.TestCase):
    def setUp(self):
        directory = os.path.dirname(os.path.realpath(__file__)) + "/resources/"
        self.input_data = open(directory + "input.conll", "r")

    def test_conll_reader(self):
        corpus = Corpus.from_file("test", self.input_data)
        self.assertEqual(5, len(corpus.documents))

if __name__ == '__main__':
    unittest.main()



""" Cost functions used during learning of coreference predictors. """

__author__ = 'martscsn'


def cost_based_on_consistency(arc, label="+"):
    """ Assign cost to arcs based on consistency of decision and anaphoricity.

    An anaphor-antecedent decision is consistent if either
        (a) the mentions are coreferent, or
        (b) the antecedent is the dummy mention, and the anaphor does not have
            any preceding coreferent mention among all extracted mentions.

    Note that (b) also contains cases where the mention has an antecedent in the
    gold data, but we were unable to extract this antecedent due to errors in
    mention detection.

    If the anaphor-antecedent decision represented by ``arc``is consistent, it
    gets cost 0. If the the decision is not consistent, and the antecedent is
    the dummy mention, it gets cost 2. Otherwise, it gets cost 1.

    Args:
        arc ((Mention, Mention)): A pair of mentions.
        label (str): The label to predict for the arc. Defaults to '+'.

    Return:
        (int): The cost of predicting the arc.
    """
    ana, ante = arc

    consistent = ana.decision_is_consistent(ante)

    # false new
    if not consistent and ante.is_dummy():
        return 2
    # wrong link
    elif not consistent:
        return 1
    else:
        return 0


def null_cost(arc, label="+"):
    """ Dummy cost function which always returns 0 (corresponding to not using
    a cost function at all).

    Args:
        arc ((Mention, Mention)): A pair of mentions.
        label (str): The label to predict for the arc. Defaults to '+'

    Return:
        0
    """
    return 0


""" Manage individual documents of text collections.

In particular, classes in this module allow to access the linguistic
information present in documents.
"""

from collections import defaultdict
import logging

from cort.core import mentions
from cort.core import spans

import StanfordDependencies

import nltk

logger = logging.getLogger(__name__)


__author__ = 'smartschat'


class Document(object):
    """Represents a document.

    Attributes:
        identifier (str): A unique identifier for the document.
        in_sentence_ids (list(int)): In-sentence indicies of all tokens in the
            document, for example [0, 1, 2, 0, 1, 2, 3, 4, ...]
        tokens (list(str)): All tokens.
        pos (list(str)): All part-of-speech tags.
        ner (list(str)): All named entity tags (if a token does not have a
            tag, the tag is set to NONE).
        parse (list(nltk:ParentedTree)): All parse trees.
        dep (list(list(StanfordDependencies.CoNLL.Token)): All dependencies
            represented as lists of tokens with label information and pointers
            to heads. One list for each sentence.
        speakers (list(str)): All speaker ids,
        coref (dict(span, int)): A mapping of mention spans to their
            coreference set id.
        annotated_mentions list(Mention): All annotated mentions.
        system_mentions list(Mention): The system mentions (initially empty).
    """
    def __init__(self, identifier, sentences, coref):
        """ Construct a document from sentence and coreference information.

        Args:
            identifier (str): A unique identifier for the document.
            sentences(list): A list of sentence information. The ith item
                contains information about the ith sentence. We assume that
                each ``sentences[i]`` is a 6-tuple
                ``tokens, pos, ner, speakers, parse, dep``, where

                * tokens (list(str)): All tokens in the sentence.
                * pos (list(str)): All part-of-speech tags in the sentence.
                * ner (list(str)): All named entity tags in the sentence (if a
                  token does not have a tag, the tag is set to NONE).
                * speakers (list(str)): All speaker ids in the sentence.
                * parse (str): A string representation of the sentence's parse
                  tree (should be readable by nltk)
                * dep (list(StanfordDependencies.CoNLL.Token): All dependencies
                  in the sentence represented as lists of tokens with label
                  information and pointers to heads.
            coref (dict(span, int)): A mapping of mention spans to their
            coreference set id.
        """
        self.identifier = identifier

        self.in_sentence_ids = []
        self.sentence_spans = []
        self.tokens = []
        self.pos = []
        self.ner = []
        self.parse = []
        self.dep = []
        self.speakers = []
        self.coref = coref

        for sentence in sentences:
            tokens, pos, ner, speakers, parse, dep = sentence

            offset = len(self.tokens)

            self.in_sentence_ids += list(range(0, len(tokens)))

            self.sentence_spans.append(spans.Span(
                offset, offset + len(tokens) - 1
            ))

            self.tokens += tokens
            self.pos += pos
            self.ner += ner
            self.parse.append(nltk.ParentedTree.fromstring(parse))
            self.dep.append(dep)
            self.speakers += speakers

        self.annotated_mentions = self.__get_annotated_mentions()
        self.system_mentions = []

    def __get_annotated_mentions(self):
        mention_spans = sorted(list(self.coref.keys()))

        seen = set()

        annotated_mentions = []

        for span in mention_spans:
            set_id = self.coref[span]
            annotated_mentions.append(
                mentions.Mention.from_document(
                    span, self, first_in_gold_entity=set_id not in seen
                )
            )
            seen.add(set_id)

        return annotated_mentions

    def __repr__(self):
        return self.identifier

    def __hash__(self):
        return hash(self.identifier)

    def __lt__(self, other):
        """ Check whether this document is less than another document.

        It is less than another document if and only if
        ``(self.folder, self.id, self.part) <
        (another.folder, another.id, another.part)``,
        where the folders and ids are compared lexicographically.

        Args:
            other (ConLLDocument): A document.

        Returns:
            True if this document is less than other, False otherwise.
        """
        return self.identifier < other.identifier

    def __eq__(self, other):
        """ Check for document equality.

        Two documents are considered equal is they have the same folder, id and
        the part.

        Args:
            other (ConLLDocument): A document.

        Returns:
            True if id and part of the documents are equal.
        """
        if isinstance(other, self.__class__):
            return self.identifier == other.identifier
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def write_antecedent_decisions_to_file(self, file):
        """ Write all antecedent decisions to a file.

        One decision is represented as one line in the file in the following
        format:

        ``identifier    anaphor_span    antecedent_span``

        For example ``(bc/cctv/00/cctv_0000); part 000   (10, 11) (1, 1)``

        Args:
            file (file): The file to write antecedent decisions to.
        """

        for mention in self.system_mentions:
            ante = mention.attributes["antecedent"]

            if ante:
                file.write(self.identifier + "\t" +
                           str(mention.span) + "\t" +
                           str(ante.span) + "\n")

    def get_annotated_mentions_from_antecedent_decisions(self, span_pairs):
        """ Overwrite coreference attributes with information from span pairs.

        In particular, interpret the spans in span pairs as anaphor/antededent
        decisions and update attributes annotated_mentions,
        spans_to_annotated_mentions, coref accordingly.

        Args:
            span_pairs (list((Span,Span))): A list of span pairs corresponding
                to anaphor/antecedent decisions.
        """

        spans_to_annotated_mentions = {}

        for mention in self.annotated_mentions:
            spans_to_annotated_mentions[mention.span] = mention

        set_id = 0
        for span_anaphor, span_antecedent in span_pairs:
            if span_antecedent not in spans_to_annotated_mentions:
                spans_to_annotated_mentions[span_antecedent] = \
                    mentions.Mention.from_document(span_antecedent, self)
                spans_to_annotated_mentions[
                    span_antecedent].attributes["annotated_set_id"] = set_id
                set_id += 1
            if span_anaphor not in spans_to_annotated_mentions:
                spans_to_annotated_mentions[span_anaphor] = \
                    mentions.Mention.from_document(span_anaphor, self)
                spans_to_annotated_mentions[span_anaphor].attributes[
                    "annotated_set_id"] = spans_to_annotated_mentions[
                        span_antecedent].attributes["annotated_set_id"]

            spans_to_annotated_mentions[span_anaphor].attributes[
                "antecedent"] = spans_to_annotated_mentions[
                    span_antecedent]

            self.annotated_mentions = sorted(
                list(spans_to_annotated_mentions.values()))

            self.coref.clear()

            for span in spans_to_annotated_mentions:
                self.coref[span] = spans_to_annotated_mentions[
                    span].attributes["annotated_set_id"]

    def get_antecedent_decisions(self, which_mentions="annotated"):
        """ Get all antecedent decisions in this document.

        Args:
            which_mentions (str): Either "annotated" or "system". Defaults to
                "system". Signals whether to consider annotated mentions or
                system mentions.

        Returns:
            dict(Mention, Mention): The mapping of antecedent decisions.
        """
        antecedent_decisions = {}

        doc_mentions = None

        if which_mentions == "annotated":
            doc_mentions = self.annotated_mentions
        elif which_mentions == "system":
            doc_mentions = self.system_mentions

        for mention in doc_mentions:
            antecedent = mention.attributes["antecedent"]

            if antecedent:
                antecedent_decisions[mention] = antecedent

        return antecedent_decisions

    def get_sentence_id_and_span(self, span):
        """ Get the sentence span from the sentence embedding the span.

        Args:
            span (Span): A span corresponding to a fragment of the document.

        Returns:
            Span: The span of the sentence which embeds the text corresponding
            to the span.
        """
        for i, sentence_span in enumerate(self.sentence_spans):
            if sentence_span.embeds(span):
                return i, sentence_span

    def to_simple_output(self):
        """ Convert the document into a simple textual representation,
        containing tokens and coreference information.

        In particular, the representation has one sentence per line, one space
        between each token. Each system mention is enclosed by
        ``<mention> ... </mention>`` tags, with the following attributes:

        * id (int): integer identifier of the mention. Equals position in the
            list of system mentions (excluding the dummy mention)
        * span_start (int): start of the mention span
        * span_end (int): end of the mention span
        * entity (int): set id of the mention. Only when the mention is in some
            coreference cluster
        * antecedent (int): id of the antecedent. Only if the mention has some
            antecedent.

        Returns:
            (str): A textual representation of document as described above.
        """
        content = []

        for sentence_span in self.sentence_spans:
            content += self.tokens[sentence_span.begin:
                                   sentence_span.end+1]

        mention_to_id = {}

        for i, mention in enumerate(self.system_mentions[1:]):
            mention_to_id[mention] = i

            tag_begin = "<mention id=\"" + str(i) + "\" " \
                        "span_start=\"" + str(mention.span.begin) + "\" " \
                        "span_end=\"" + str(mention.span.end) + "\""

            if mention.attributes["set_id"]:
                tag_begin += " entity=\"" + \
                             str(mention.attributes["set_id"]) + "\""

            if mention.attributes["antecedent"]:
                antecedent_id = mention_to_id[
                    mention.attributes["antecedent"]]

                tag_begin += " antecedent=\"" + str(antecedent_id) + "\""

            tag_begin += ">"

            old_begin = content[mention.span.begin]
            content[mention.span.begin] = tag_begin + old_begin

            content[mention.span.end] += "</mention>"

        output_string = ""

        for sentence_span in self.sentence_spans:
            output_string += " ".join(
                content[sentence_span.begin:sentence_span.end+1]) + "\n"

        return output_string

    def get_html_friendly_identifier(self):
        """ Transform the identifier to a HTML/CSS/JS-friendly representation
        for visualization.

        Returns:
            str: The HTML/CSS-JS-friendly representation.
        """
        return self.identifier.replace(".", "_").replace("/", "_")


class CoNLLDocument(Document):
    """Represents a document in CoNLL format.

    For a specification of the format, see
    http://conll.cemantix.org/2012/data.html.

    Attributes:
        identifier (str): A unique identifier for the document.
        in_sentence_ids (list(int)): In-sentence indicies of all tokens in the
            document, for example [0, 1, 2, 0, 1, 2, 3, 4, ...]
        tokens (list(str)): All tokens.
        pos (list(str)): All part-of-speech tags.
        ner (list(str)): All named entity tags (if a token does not have a
            tag, the tag is set to NONE).
        parse (list(nltk:ParentedTree)): All parse trees.
        dep (list(list(StanfordDependencies.CoNLL.Token)): All dependencies
            represented as lists of tokens with label information and pointers
            to heads. One list for each sentence.
        speakers (list(str)): All speaker ids,
        coref (dict(span, int)): A mapping of mention spans to their
            coreference set id.
        annotated_mentions list(Mention): All annotated mentions.
        system_mentions list(Mention): The system mentions (initially empty).
        document_table (list(list(str))): A tabular representation of the
            document (as in the CoNLL data).
    """

    def __init__(self, document_as_string):
        """ Construct a document from a string representation.

            The Format must follow the CoNLL format, see
                http://conll.cemantix.org/2012/data.html.

            Args:
                document_as_string (str): A representation of a document in
                    the CoNLL format.
            """
        identifier = " ".join(document_as_string.split("\n")[0].split(" ")[2:])

        self.document_table = CoNLLDocument.__string_to_table(
            document_as_string)
        in_sentence_ids = [int(i) for i in self.__extract_from_column(2)]
        indexing_start = in_sentence_ids[0]
        if indexing_start != 0:
            logger.warning("Detected " +
                           str(indexing_start) +
                           "-based indexing for tokens in sentences in input,"
                           "transformed to 0-based indexing.")
            in_sentence_ids = [i - indexing_start for i in in_sentence_ids]
        sentence_spans = CoNLLDocument.__extract_sentence_spans(in_sentence_ids)
        temp_tokens = self.__extract_from_column(3)
        temp_pos = self.__extract_from_column(4)
        temp_ner = self.__extract_ner()
        temp_speakers = self.__extract_from_column(9)
        coref = CoNLLDocument.__get_span_to_id(self.__extract_from_column(-1))
        parses = [CoNLLDocument.get_parse(span,
                                          self.__extract_from_column(5),
                                          temp_pos,
                                          temp_tokens)
                  for span in sentence_spans]
        sd = StanfordDependencies.get_instance()
        dep_trees = sd.convert_trees(
            [parse.replace("NOPARSE", "S") for parse in parses],
        )
        sentences = []
        for i, span in enumerate(sentence_spans):
            sentences.append(
                (temp_tokens[span.begin:span.end + 1],
                 temp_pos[span.begin:span.end + 1],
                 temp_ner[span.begin:span.end + 1],
                 temp_speakers[span.begin:span.end + 1],
                 parses[i],
                 dep_trees[i])
            )

        super(CoNLLDocument, self).__init__(identifier, sentences, coref)

    def __extract_from_column(self, column):
        entries = []
        for line in self.document_table:
            entries.append(line[column])

        return entries

    def __extract_ner(self):
        entries = self.__extract_from_column(10)
        ner = []

        tag = "NONE"
        for i in range(0, len(entries)):
            entry = entries[i]

            if "(" in entry:
                tag = entry.strip("(").strip(")").strip("*")

            ner.append(tag)

            if ")" in entry:
                tag = "NONE"

        return ner

    @staticmethod
    def __string_to_table(document_as_string):
        table = []

        document_contents = document_as_string.split("\n")[1:-2]

        for line in document_contents:
            if line != "" and not line.isspace():
                table.append(line.split())

        return table

    @staticmethod
    def __extract_sentence_spans(in_sentence_ids):
        sentence_spans = []

        span_start = 0

        for i in range(1, len(in_sentence_ids)):
            if in_sentence_ids[i] <= in_sentence_ids[i-1]:
                sentence_spans.append(spans.Span(span_start, i-1))
                span_start = i

        sentence_spans.append(spans.Span(span_start,
                                         len(in_sentence_ids)-1))

        return sentence_spans

    @staticmethod
    def __get_span_to_id(column):
        span_to_id = {}

        ids_to_stack = defaultdict(list)

        for i in range(0, len(column)):
            entry = column[i]

            if entry != "-":
                parallel_annotations = entry.split("|")

                for annotation in parallel_annotations:
                    if annotation.startswith("(") and annotation.endswith(")"):
                        set_id = annotation[1:-1]
                        span_to_id[spans.Span(i, i)] = int(set_id)
                    elif annotation.startswith("("):
                        set_id = annotation[1:]
                        ids_to_stack[set_id].append(i)
                    elif annotation.endswith(")"):
                        set_id = annotation[:-1]
                        span_to_id[
                            spans.Span(ids_to_stack[set_id].pop(), i)
                        ] = int(set_id)

        return span_to_id

    @staticmethod
    def get_parse(span, parses_as_string, pos, tokens):
        """ Get a the parse tree (as a string) of to the span.

        Args:
            span (Span): A span corresponding to a fragment of the document.

        Returns:
            str: A string representation of the parse tree of the span.
        """
        parse_tree = ""
        for i in range(span.begin, span.end+1):
            parse_bit = parses_as_string[i]
            parse_tree += \
                parse_bit.replace("(", " (").replace(
                    "*", " (" + pos[i] + " " + tokens[i] + ")")

        return parse_tree.strip()

    def get_string_representation(self):
        """ Get a string representation of the document.

        Returns:
            str: A string representation of the document which conforms to the
            CoNLL format specifications
            (http://conll.cemantix.org/2012/data.html).
        """
        mention_string_representation = \
            CoNLLDocument.__get_string_representation_of_mentions(
                len(self.document_table), self.system_mentions)

        new_table = self.document_table

        for row, mention_row in zip(new_table, mention_string_representation):
            row[-1] = mention_row

        padded_table = []

        current_row = new_table[0]
        padded_table.append(current_row)

        for i in range(1, len(new_table)):
            previous_row = current_row
            current_row = new_table[i]

            if int(current_row[2]) <= int(previous_row[2]):
                padded_table.append("")

            padded_table.append(current_row)

        begin = ("#begin document " + self.identifier + "\n")

        content = "\n".join(["\t".join(row) for row in padded_table])

        end = "\n#end document\n"

        return begin + content + end

    @staticmethod
    def __get_string_representation_of_mentions(length, mentions_in_doc):
        index_to_strings = defaultdict(list)

        for mention in mentions_in_doc:
            set_id = mention.attributes["set_id"]

            if set_id is None:
                continue

            span = mention.span

            if span.begin == span.end:
                index_to_strings[span.begin].append("(" + str(set_id) + ")")
            else:
                index_to_strings[span.begin].append("(" + str(set_id))
                index_to_strings[span.end].append(str(set_id) + ")")

        output_with_parallel_annotations = []

        for i in range(0, length+1):
            if i in index_to_strings:
                output_with_parallel_annotations.append(
                    "|".join(sorted(index_to_strings[i])))
            else:
                output_with_parallel_annotations.append("-")

        return output_with_parallel_annotations

    def get_html_friendly_identifier(self):
        """ Transform the identifier to a HTML/CSS/JS-friendly representation
        for visualization.

        Returns:
            str: The HTML/CSS/JS-friendly representation.
        """
        splitted_by_whitespace = self.identifier.split()
        return splitted_by_whitespace[0].split("/")[-1][:-2] + \
               "_part_" + splitted_by_whitespace[-1]



import unittest

from cort.core.mentions import Mention
from cort.core.spans import Span
from cort.core.documents import CoNLLDocument


__author__ = 'smartschat'


class TestDiscourseUnits(unittest.TestCase):
    def setUp(self):
        self.real_example = """#begin document (bn/voa/02/voa_0220); part 000
bn/voa/02/voa_0220   0    0    Unidentified    JJ  (TOP(S(NP(NP*          -   -   -   -            *    -
bn/voa/02/voa_0220   0    1          gunmen   NNS              *)         -   -   -   -            *    -
bn/voa/02/voa_0220   0    2              in    IN           (PP*          -   -   -   -            *    -
bn/voa/02/voa_0220   0    3           north    JJ      (NP(ADJP*          -   -   -   -            *    -
bn/voa/02/voa_0220   0    4         western    JJ              *)         -   -   -   -            *    -
bn/voa/02/voa_0220   0    5        Colombia   NNP            *)))         -   -   -   -         (GPE)   -
bn/voa/02/voa_0220   0    6            have   VBP           (VP*        have  -   -   -            *    -
bn/voa/02/voa_0220   0    7       massacred   VBN           (VP*    massacre  -   -   -            *    -
bn/voa/02/voa_0220   0    8              at    IN   (NP(QP(ADVP*          -   -   -   -   (CARDINAL*    -
bn/voa/02/voa_0220   0    9           least   JJS              *)         -   -   -   -            *    -
bn/voa/02/voa_0220   0   10          twelve    CD              *)         -   -   -   -            *)   -
bn/voa/02/voa_0220   0   11        peasants   NNS              *)         -   -   -   -            *    -
bn/voa/02/voa_0220   0   12              in    IN           (PP*          -   -   -   -            *    -
bn/voa/02/voa_0220   0   13             the    DT        (NP(NP*          -   -   -   -            *   (0
bn/voa/02/voa_0220   0   14          second    JJ              *          -   -   -   -     (ORDINAL)   -
bn/voa/02/voa_0220   0   15            such    JJ              *          -   -   -   -            *    -
bn/voa/02/voa_0220   0   16        incident    NN              *)   incident  -   2   -            *    -
bn/voa/02/voa_0220   0   17              in    IN           (PP*          -   -   -   -            *    -
bn/voa/02/voa_0220   0   18              as    RB        (NP(QP*          -   -   -   -       (DATE*    -
bn/voa/02/voa_0220   0   19            many    JJ              *)         -   -   -   -            *    -
bn/voa/02/voa_0220   0   20            days   NNS         *))))))        day  -   4   -            *)   0)
bn/voa/02/voa_0220   0   21               .     .             *))         -   -   -   -            *    -

bn/voa/02/voa_0220   0    0          Local    JJ    (TOP(S(NP*          -    -   -   -   *   (ARG0*             *    -
bn/voa/02/voa_0220   0    1         police   NNS             *)     police   -   -   -   *        *)            *    -
bn/voa/02/voa_0220   0    2            say   VBP          (VP*         say  01   1   -   *      (V*)            *    -
bn/voa/02/voa_0220   0    3             it   PRP   (SBAR(S(NP*)         -    -   -   -   *   (ARG1*        (ARG1*)   -
bn/voa/02/voa_0220   0    4             's   VBZ          (VP*          be  01   1   -   *        *           (V*)   -
bn/voa/02/voa_0220   0    5            not    RB             *          -    -   -   -   *        *    (ARGM-NEG*)   -
bn/voa/02/voa_0220   0    6          clear    JJ        (ADJP*)         -    -   -   -   *        *        (ARG2*)   -
bn/voa/02/voa_0220   0    7            who    WP   (SBAR(WHNP*)         -    -   -   -   *        *             *    -
bn/voa/02/voa_0220   0    8            was   VBD        (S(VP*          be   -   1   -   *        *             *    -
bn/voa/02/voa_0220   0    9    responsible    JJ        (ADJP*          -    -   -   -   *        *             *    -
bn/voa/02/voa_0220   0   10            for    IN          (PP*          -    -   -   -   *        *             *    -
bn/voa/02/voa_0220   0   11            the    DT          (NP*          -    -   -   -   *        *             *   (0
bn/voa/02/voa_0220   0   12       massacre    NN    *))))))))))   massacre   -   -   -   *        *)            *    0)
bn/voa/02/voa_0220   0   13              .     .            *))         -    -   -   -   *        *             *    -

#end document
"""

        self.real_document = CoNLLDocument(self.real_example)

        self.complicated_mention_example = """#begin	document	(test2);	part	000
test2	0	0	This    NN   (NP*	-   -   -   -   -   (0)
test2	0	1	is  NN	*   -   -   -   -   -   -
test2	0	2	just    NN   *	-   -   -   -   -   -
test2	0	3	a   NN   *	-   -   -   -   -   (0|(1)
test2	0	4	test    NN   *	-   -   -   -   -   0)
test2	0	5	.   NN   *)	-   -   -   -   -   -

test2	0	0	It  NN   (NP*	-   -   -   -   -   (1)|(4
test2	0	1	shows   NN   *	-   -   -   -   -   -
test2	0	2	that    NN   *	-   -   -   -   -   (2)
test2	0	3	the NN   *	-   -   -   -   -   (2|(3
test2	0	4	scorer  NN   *	-   -   -   -   -   2)|4)
test2	0	5	works   NN   *	-   -   -   -   -   3)
test2	0	6	.   NN   *)	-   -   -   -   -   -

#end	document"""

        self.complicated_mention_document = CoNLLDocument(self.complicated_mention_example)

        self.for_head_example = """#begin document (wb/a2e/00/a2e_0000); part 000
wb/a2e/00/a2e_0000      0       0       Celebration     NN      (TOP(S(NP*      -       -       -       -       *       (ARG0*  -
wb/a2e/00/a2e_0000      0       1       Shooting        NN      *)      shoot   -       -       -       *       *)      -
wb/a2e/00/a2e_0000      0       2       Turns   VBZ     (VP*    turn    02      2       -       *       (V*)    -
wb/a2e/00/a2e_0000      0       3       Wedding NN      (NP*)   wed     -       -       -       *       (ARG1*) (3)
wb/a2e/00/a2e_0000      0       4       Into    IN      (PP*    -       -       -       -       *       (ARG2*  -
wb/a2e/00/a2e_0000      0       5       a       DT      (NP*    -       -       -       -       *       *       -
wb/a2e/00/a2e_0000      0       6       Funeral NN      *)      -       -       -       -       *       *       -
wb/a2e/00/a2e_0000      0       7       in      IN      (PP*    -       -       -       -       *       *       -
wb/a2e/00/a2e_0000      0       8       Southern        JJ      (NP*    -       -       -       -       *       *       (14
wb/a2e/00/a2e_0000      0       9       Gaza    NNP     *       -       -       -       -       (GPE*   *       -
wb/a2e/00/a2e_0000      0       10      Strip   NNP     *)))))) -       -       -       -       *)      *)      14)

#end document"""

        self.for_head_document = CoNLLDocument(self.for_head_example)

        self.date_mention_example = """#begin document (nw/wsj/24/wsj_2444); part 000
nw/wsj/24/wsj_2444   0   0    Employment    NN      (TOP(S(NP*)   employment  01   1   -           *    (V*)       (ARG1*)     -
nw/wsj/24/wsj_2444   0   1            is   VBZ            (VP*            be  01   1   -           *      *           (V*)     -
nw/wsj/24/wsj_2444   0   2           now    RB          (ADVP*)           -    -   -   -           *      *    (ARGM-TMP*)     -
nw/wsj/24/wsj_2444   0   3            4     CD  (ADJP(ADJP(QP*            -    -   -   -   (PERCENT*      *        (ARG2*      -
nw/wsj/24/wsj_2444   0   4             %    NN               *)           -    -   -   -           *)     *             *      -
nw/wsj/24/wsj_2444   0   5        higher   JJR               *)           -    -   -   -           *      *             *      -
nw/wsj/24/wsj_2444   0   6          than    IN            (PP*            -    -   -   -           *      *             *      -
nw/wsj/24/wsj_2444   0   7            in    IN            (PP*            -    -   -   -           *      *             *      -
nw/wsj/24/wsj_2444   0   8         1983     CD        (NP*)))))           -    -   -   -       (DATE)     *             *)   (16)
nw/wsj/24/wsj_2444   0   9             .     .              *))           -    -   -   -           *      *             *      -

bc/cctv/00/cctv_0000   0    0           For     IN  (TOP(S(PP*          -    -   -   Speaker#1        *      *    (ARGM-TMP*      -
bc/cctv/00/cctv_0000   0    1           two     CD        (NP*          -    -   -   Speaker#1   (DATE*      *             *      (1
bc/cctv/00/cctv_0000   0    2         years    NNS          *))         -    -   -   Speaker#1        *)     *             *)     1)
bc/cctv/00/cctv_0000   0    3             ,      ,           *          -    -   -   Speaker#1        *      *             *      -
bc/cctv/00/cctv_0000   0    4        Disney    NNP        (NP*)         -    -   -   Speaker#1     (ORG)     *        (ARG0*)   (12)
bc/cctv/00/cctv_0000   0    5           has    VBZ        (VP*        have  01   -   Speaker#1        *    (V*)            *      -
bc/cctv/00/cctv_0000   0    6    constantly     RB      (ADVP*)         -    -   -   Speaker#1        *      *    (ARGM-MNR*)     -
bc/cctv/00/cctv_0000   0    7    maintained    VBN        (VP*    maintain  01   1   Speaker#1        *      *           (V*)     -
bc/cctv/00/cctv_0000   0    8           its   PRP$        (NP*          -    -   -   Speaker#1        *      *        (ARG1*    (12)
bc/cctv/00/cctv_0000   0    9       mystery     NN         *)))         -    -   -   Speaker#1        *      *             *)     -
bc/cctv/00/cctv_0000   0   10             .      .          *))         -    -   -   Speaker#1        *      *             *      -

nw/wsj/24/wsj_2413   0    0    Government    NNP    (TOP(S(NP(NP*          -    -   -   -        *     (ARG0*        *       (ARG0*    (ARG0*         *    (16
nw/wsj/24/wsj_2413   0    1     officials    NNS                *)   official   -   1   -        *          *        *            *         *         *      -
nw/wsj/24/wsj_2413   0    2          here     RB       (UCP(ADVP*)         -    -   -   -        *          *        *            *         *         *      -
nw/wsj/24/wsj_2413   0    3           and     CC                *          -    -   -   -        *          *        *            *         *         *      -
nw/wsj/24/wsj_2413   0    4            in     IN             (PP*          -    -   -   -        *          *        *            *         *         *      -
nw/wsj/24/wsj_2413   0    5         other     JJ             (NP*          -    -   -   -        *          *        *            *         *         *      -
nw/wsj/24/wsj_2413   0    6     countries    NNS             *))))    country   -   3   -        *          *)       *            *)        *)        *     16)
nw/wsj/24/wsj_2413   0    7          laid    VBD          (VP(VP*         lay  01   2   -        *        (V*)       *            *         *         *      -
nw/wsj/24/wsj_2413   0    8         plans    NNS          (NP(NP*)       plan   -   2   -        *     (ARG1*        *            *         *         *      -
nw/wsj/24/wsj_2413   0    9       through     IN             (PP*          -    -   -   -        *          *        *            *         *         *      -
nw/wsj/24/wsj_2413   0   10           the     DT             (NP*          -    -   -   -   (DATE*          *        *            *         *         *     (6
nw/wsj/24/wsj_2413   0   11       weekend     NN               *))    weekend   -   -   -        *)         *        *            *         *         *      6)
nw/wsj/24/wsj_2413   0   12            to     TO           (S(VP*          -    -   -   -        *   (C-ARG1*        *            *         *         *      -
nw/wsj/24/wsj_2413   0   13          head     VB             (VP*        head  03   6   -        *          *      (V*)           *         *         *      -
nw/wsj/24/wsj_2413   0   14           off     RP            (PRT*)         -    -   -   -        *          *        *            *         *         *      -
nw/wsj/24/wsj_2413   0   15             a     DT             (NP*          -    -   -   -        *          *   (ARG1*            *         *         *      -
nw/wsj/24/wsj_2413   0   16        Monday    NNP                *          -    -   -   -    (DATE)         *        *            *         *         *     (8)
nw/wsj/24/wsj_2413   0   17        market     NN                *      market   -   4   -        *          *        *            *         *         *      -
nw/wsj/24/wsj_2413   0   18      meltdown     NN           *))))))         -    -   -   -        *         *))       *)           *         *         *      -
nw/wsj/24/wsj_2413   0   19           --       :                *          -    -   -   -        *          *        *            *         *         *      -
nw/wsj/24/wsj_2413   0   20           but     CC                *          -    -   -   -        *          *        *            *         *         *      -
nw/wsj/24/wsj_2413   0   21          went    VBD             (VP*          go  02   1   -        *          *        *          (V*)        *         *      -
nw/wsj/24/wsj_2413   0   22           out     IN             (PP*          -    -   -   -        *          *        *   (ARGM-DIR*         *         *      -
nw/wsj/24/wsj_2413   0   23            of     IN             (PP*          -    -   -   -        *          *        *            *         *         *      -
nw/wsj/24/wsj_2413   0   24         their   PRP$             (NP*          -    -   -   -        *          *        *            *         *         *    (16)
nw/wsj/24/wsj_2413   0   25           way     NN              *)))        way   -   5   -        *          *        *            *)        *         *      -
nw/wsj/24/wsj_2413   0   26            to     TO           (S(VP*          -    -   -   -        *          *        *   (ARGM-PNC*         *         *      -
nw/wsj/24/wsj_2413   0   27          keep     VB             (VP*        keep  02   1   -        *          *        *            *       (V*)        *      -
nw/wsj/24/wsj_2413   0   28         their   PRP$             (NP*          -    -   -   -        *          *        *            *         *    (ARG0*)   (16)
nw/wsj/24/wsj_2413   0   29         moves    NNS                *)       move  02   2   -        *          *        *            *         *       (V*)     -
nw/wsj/24/wsj_2413   0   30         quiet     JJ   (S(ADJP*)))))))         -    -   -   -        *          *        *            *)   (ARG1*)        *      -
nw/wsj/24/wsj_2413   0   31             .      .               *))         -    -   -   -        *          *        *            *         *         *      -

#end document"""

        self.date_mention_document = CoNLLDocument(self.date_mention_example)

    def test_mention_tokens(self):
        expected = ["the", "massacre"]
        self.assertEqual(
            expected,
            Mention.from_document(
                Span(33, 34),
                self.real_document).attributes["tokens"])

        expected = "the massacre"
        self.assertEqual(
            expected,
            Mention.from_document(
                Span(33, 34),
                self.real_document).attributes["tokens_as_lowercase_string"])

    def test_mention_type(self):
        self.assertEqual(
            "NAM",
            Mention.from_document(
                Span(37, 37),
                self.date_mention_document).attributes["type"])
        self.assertEqual(
            "NAM",
            Mention.from_document(
                Span(11, 12),
                self.date_mention_document).attributes["type"])
        self.assertEqual(
            "NAM",
            Mention.from_document(
                Span(31, 32),
                self.date_mention_document).attributes["type"])
        self.assertEqual(
            "NAM",
            Mention.from_document(
                Span(8, 8),
                self.date_mention_document).attributes["type"])
        self.assertEqual(
            "NOM",
            Mention.from_document(
                Span(33, 34),
                self.real_document).attributes["type"])

    def test_mention_set_id(self):
        self.assertEqual(
            0,
            Mention.from_document(
                Span(33, 34),
                self.real_document).attributes["annotated_set_id"])
        self.assertEqual(
            4,
            Mention.from_document(
                Span(6, 10),
                self.complicated_mention_document).attributes[
                    "annotated_set_id"])
        self.assertEqual(
            1,
            Mention.from_document(
                Span(3, 3),
                self.complicated_mention_document).attributes[
                    "annotated_set_id"])

    def test_mention_get_head(self):
        expected = ["massacre"]
        self.assertEqual(
            expected,
            Mention.from_document(
                Span(33, 34),
                self.real_document).attributes["head"])

        expected = ["Wedding"]
        self.assertEqual(
            expected,
            Mention.from_document(
                Span(3, 3),
                self.for_head_document).attributes["head"])

        expected = "wedding"
        self.assertEqual(
            expected,
            Mention.from_document(
                Span(3, 3),
                self.for_head_document).attributes["head_as_lowercase_string"])

    def test_mention_get_governor(self):
        expected = "massacred"
        self.assertEqual(
            expected,
            Mention.from_document(
                Span(0, 1),
                self.real_document).attributes["governor"])

    def test_mention_get_ancestry(self):
        expected = "-L-VBN-L-NONE"
        self.assertEqual(
            expected,
            Mention.from_document(
                Span(11, 11),
                self.real_document).attributes["ancestry"])

        expected = "-R-NNS-R-VBN"
        self.assertEqual(
            expected,
            Mention.from_document(
                Span(0, 0),
                self.real_document).attributes["ancestry"])

    def test_mention_get_head_span(self):
        self.assertEqual(
            Span(9, 10),
            Mention.from_document(
                Span(8, 10),
                self.for_head_document).attributes["head_span"])

    def test_mention_get_fine_type(self):
        self.assertEqual(
            "DEF",
            Mention.from_document(
                Span(33, 34),
                self.real_document).attributes["fine_type"])
        self.assertEqual(
            "DEF",
            Mention.from_document(
                Span(21, 27),
                self.date_mention_document).attributes["fine_type"])
        self.assertEqual(
            "INDEF",
            Mention.from_document(
                Span(22, 22),
                self.date_mention_document).attributes["fine_type"])
        self.assertEqual(
            "POSS_ADJ",
            Mention.from_document(
                Span(45, 45),
                self.date_mention_document).attributes["fine_type"])

    def test_mention_get_sentence_id(self):
        self.assertEqual(
            0,
            Mention.from_document(
                Span(13, 20),
                self.real_document).attributes["sentence_id"])
        self.assertEqual(
            1,
            Mention.from_document(
                Span(33, 34),
                self.real_document).attributes["sentence_id"])

    def test_mention_get_context(self):
        self.assertEqual(
            ["laid", "plans"],
            Mention.from_document(
                Span(21, 27),
                self.date_mention_document).get_context(2))
        self.assertEqual(
            None,
            Mention.from_document(
                Span(21, 27),
                self.date_mention_document).get_context(0))
        self.assertEqual(
            ["through"],
            Mention.from_document(
                Span(31, 32),
                self.date_mention_document).get_context(-1))
        self.assertEqual(
            None,
            Mention.from_document(
                Span(21, 27),
                self.date_mention_document).get_context(1000))

    def test_is_coreferent_with(self):

        self.assertEqual(True,
            Mention(
                None, Span(0, 0), {"annotated_set_id": 1}
            ).is_coreferent_with(
                Mention(None, Span(3, 4), {"annotated_set_id": 1})
            )
        )

        self.assertEqual(False,
            Mention(
                None, Span(0, 0), {"annotated_set_id": 1}
            ).is_coreferent_with(
                Mention(None, Span(3, 4), {"annotated_set_id": 0})
            )
        )

        self.assertEqual(False,
            Mention(
                None, Span(0, 0), {"annotated_set_id": None}
            ).is_coreferent_with(
                Mention(None, Span(3, 4), {"annotated_set_id": None})
            )
        )

        self.assertEqual(True,
            Mention(
                self.complicated_mention_document, Span(0, 0), {"annotated_set_id": 1}
            ).is_coreferent_with(
                Mention(self.complicated_mention_document, Span(3, 4), {"annotated_set_id": 1})
            )
        )

        self.assertEqual(False,
            Mention(
                self.complicated_mention_document, Span(0, 0),
                {"annotated_set_id": None}
            ).is_coreferent_with(
                Mention(self.complicated_mention_document, Span(3, 4),
                        {"annotated_set_id": None})
            )
        )

        self.assertEqual(False,
            Mention(
                self.complicated_mention_document, Span(0, 0), {"annotated_set_id": 1}
            ).is_coreferent_with(
                Mention(self.real_document, Span(13, 20),
                        {"annotated_set_id": 1})
            )
        )


if __name__ == '__main__':
    unittest.main()


"""
Client and process monitor for the java socket server.
"""

from __future__ import division
import subprocess, tempfile, time, os, logging, re, struct, socket, atexit, \
    glob, itertools, codecs
from copy import copy,deepcopy
from pprint import pprint
try:
    import ujson as json
except ImportError:
    import json

# SUGGESTED: for constituent parsing models, specify shift-reduce parser in
# configdict with:
#  'parse.model': 'edu/stanford/nlp/models/srparser/englishSR.ser.gz'

MODES_items = [

('ssplit',  {'annotators': "tokenize, ssplit",
    'description': "tokenization and sentence splitting (included in all subsequent ones)", }),
('pos',     {'annotators':"tokenize, ssplit, pos, lemma",
    'description':"POS (and lemmas)",}),
('ner',     {'annotators':"tokenize, ssplit, pos, lemma, ner, entitymentions",
    'description':"POS and NER (and lemmas)",}),
('parse',    {'annotators':"tokenize, ssplit, pos, lemma, parse",
    'description':"fairly basic parsing with POS, lemmas, trees, dependencies",}),
('nerparse', {'annotators':"tokenize, ssplit, pos, lemma, ner, entitymentions, parse",
    'description':"parsing with NER, POS, lemmas, depenencies."}),
('coref', {'annotators':"tokenize, ssplit, pos, lemma, ner, entitymentions, parse, dcoref",
    'description':"Coreference, including constituent parsing."})
]

MODES = dict(MODES_items)

#logging.basicConfig()  # wtf, why we have to call this?
LOG = logging.getLogger("CoreNLP_PyWrapper")
LOG.setLevel("INFO")
# LOG.setLevel("DEBUG")

PARSEDOC_TIMEOUT_SEC = 60 * 5
STARTUP_BUSY_WAIT_INTERVAL_SEC = 1.0

def command(mode=None, configfile=None, configdict=None, comm_mode=None,
        java_command="java",
        java_options="-Xmx4g -XX:ParallelGCThreads=1",
        **kwargs):
    d = {}
    d.update(**locals())
    d.update(**kwargs)

    more_config = ""
    if mode is None and configfile is None and configdict is None:
        assert False, "Need to set mode, or the annotators directly, for this wrapper to work."
    if mode:
        if configdict is not None:
            assert 'annotators' not in configdict, "mode was given but annotators are set in the configdict.  use only one please."
        elif configdict is None:
            configdict = {}
        LOG.info("mode given as '%s' so setting annotators: %s" % (mode, MODES[mode]['annotators']))
        configdict['annotators'] = MODES[mode]['annotators']
    if configfile:
        more_config += " --configfile {}".format(configfile)
    if configdict:
        j = json.dumps(configdict)
        assert "'" not in j, "can't handle single quote in config values"
        more_config += " --configdict '{}'".format(j)
    d['more_config'] = more_config

    if comm_mode=='SOCKET':
        d['comm_info'] = "--server {server_port}".format(**d)
    elif comm_mode=='PIPE':
        d['comm_info'] = "--outpipe {outpipe}".format(**d)
    else: assert False, "need comm_mode to be SOCKET or PIPE but got " + repr(comm_mode)


    cmd = """exec {java_command} {java_options} -cp '{classpath}' 
    corenlp.SocketServer {comm_info} {more_config}"""
    return cmd.format(**d).replace("\n", " ")


class SubprocessCrashed(Exception):
    pass


class CoreNLP:

    def __init__(self, mode=None, 
            configfile=None, configdict=None,
            corenlp_jars=(
                "/home/sw/corenlp/stanford-corenlp-full-2015-04-20/*",
                "/home/sw/stanford-srparser-2014-10-23-models.jar",
                ),
            comm_mode='PIPE',  # SOCKET or PIPE
            server_port=12340, outpipe_filename_prefix="/tmp/corenlp_pywrap_pipe",
            **more_configdict_args
            ):
        """
        mode: if you supply this as a single string, we'll use a prebaked set
        of annotators.  if you don't want this, specify either (configfile or
        configdict) and set 'annotators' there.

        corenlp_jars: jars for the classpath.  tuple or list of strings.
        this is just passed on to java's "-cp" flag; note that it accepts
        limited use of wildcards.

        configfile, configdict: can give a corenlp configuration, either as an
        external file (like the .ini or java properties format, whatever it
        is), or else as a python dictionary.  we just pass it on, though we
        will look at the annotators setting.

        Extra keyword arguments are added to the configdict.
        Note you can't pass options with dots in them this way.

        server_port: have to specify this if you want to run multple instances
        in separate processes.  todo we should use some other communication
        mechanism that doesnt have to worry about this
        """
        self.mode = mode
        self.proc = None
        self.server_port = server_port
        self.configfile = configfile
        self.comm_mode = comm_mode
        self.outpipe = None

        self.configdict = deepcopy(configdict)
        if not self.configdict: self.configdict = {}
        self.configdict.update(more_configdict_args)
        if not self.configdict: self.configdict = None

        if self.comm_mode=='PIPE':
            tag = "pypid=%d_time=%s" % (os.getpid(), time.time())
            self.outpipe = "%s_%s" % (outpipe_filename_prefix, tag)
            assert not os.path.exists(self.outpipe)

        assert isinstance(corenlp_jars, (list,tuple))

        deglobbed = itertools.chain(*[glob.glob(f) for f in corenlp_jars])
        assert any(os.path.exists(f) for f in deglobbed), "CoreNLP jar files don't seem to exist; are the paths correct?  Searched files: %s" % repr(deglobbed)

        local_libdir = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                                    'lib')

        jars = [os.path.join(local_libdir, "*")]
        jars += corenlp_jars
        self.classpath = ':'.join(jars)
        # self.classpath += ":../bin:bin"  ## for eclipse java dev

        # LOG.info("CLASSPATH: " + self.classpath)

        self.start_server()
        # This probably is only half-reliable, but worth a shot.
        atexit.register(self.cleanup)

    def cleanup(self):
        self.kill_proc_if_running()
        if self.outpipe and os.path.exists(self.outpipe):
            os.unlink(self.outpipe)

    def __del__(self):
        # This is also an unreliable way to ensure the subproc is gone, but
        # might as well try
        self.cleanup()

    def start_server(self):
        self.kill_proc_if_running()

        if self.comm_mode=='PIPE':
            if not os.path.exists(self.outpipe):
                os.mkfifo(self.outpipe)
        
        cmd = command(**self.__dict__)
        LOG.info("Starting java subprocess, and waiting for signal it's ready, with command: %s" % cmd)
        self.proc = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE)
        time.sleep(STARTUP_BUSY_WAIT_INTERVAL_SEC)

        if self.comm_mode=='SOCKET':
            sock = self.get_socket(num_retries=100, retry_interval=STARTUP_BUSY_WAIT_INTERVAL_SEC)
            sock.close()
        elif self.comm_mode=='PIPE':
            self.outpipe_fp = codecs.open(self.outpipe, 'r', "ISO-8859-1")

        while True:
            # This loop is for if you have timeouts for the socket connection
            # The pipe system doesn't have timeouts, so this should run only
            # once in that case.
            try:
                ret = self.send_command_and_parse_result('PING\t""', 2)
                if ret is None:
                    continue
                assert ret == "PONG", "Bad return data on startup ping: " + ret
                LOG.info("Successful ping. The server has started.")
                break
            except socket.error as e:
                LOG.info("Waiting for startup: ping got exception: %s %s" % (type(e), e))
                LOG.info("pausing before retry")
                time.sleep(STARTUP_BUSY_WAIT_INTERVAL_SEC)

        LOG.info("Subprocess is ready.")

    def ensure_proc_is_running(self):
        if self.proc is None:
            # Has never been started
            self.start_server()
        elif self.proc.poll() is not None:
            # Restart
            self.start_server()

    def kill_proc_if_running(self):
        if self.proc is None:
            # it's never been started yet
            return
        retcode = self.proc.poll()
        if retcode is not None:
            LOG.info("Subprocess seems to be stopped, exit code %s" % retcode)
        elif retcode is None:
            LOG.warning("Killing subprocess %s" % self.proc.pid)
            os.kill(self.proc.pid, 9)

    def parse_doc(self, text, timeout=PARSEDOC_TIMEOUT_SEC, raw=False):
        cmd = "PARSEDOC\t%s" % json.dumps(text)
        return self.send_command_and_parse_result(cmd, timeout, raw=raw)

    def get_socket(self, num_retries=1, retry_interval=1):
        # could be smarter here about reusing the same socket?
        for trial in range(num_retries):
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                # sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1) # not sure if this is needed?
                sock.connect(('localhost', self.server_port))
                return sock
            except (socket.error, socket.timeout) as e:
                LOG.info("socket error when making connection (%s)" % e)
                if trial < num_retries-1:
                    LOG.info("pausing before retry")
                    time.sleep(retry_interval)
        assert False, "couldnt connect socket"

    def send_command_and_parse_result(self, cmd, timeout, raw=False):
        try:
            self.ensure_proc_is_running()
            data = self.send_command_and_get_string_result(cmd, timeout)
            if data is None: return None
            decoded = None
            if raw:
                return data
            try:
                decoded = json.loads(data)
            except ValueError:
                LOG.warning("Bad JSON returned from subprocess; returning null.")
                LOG.warning("Bad JSON length %d, starts with: %s" % (len(data), repr(data[:1000])))
                return None
            return decoded
        except socket.timeout as e:
            LOG.info("Socket timeout happened, returning None: %s %s" % (type(e), e))
            return None
            # This is tricky. maybe the process is running smoothly but just
            # taking longer than we like.  if it's in thie state, and we try to
            # send another command, what happens?  Should we forcibly restart
            # the process now just in case?

    def send_command_and_get_string_result(self, cmd, timeout):
        if self.comm_mode == 'SOCKET':
            sock = self.get_socket(num_retries=100)
            sock.settimeout(timeout)
            sock.sendall(cmd + "\n")
            size_info_str = sock.recv(8)
        elif self.comm_mode == 'PIPE':
            to_write = cmd + "\n"
            self.proc.stdin.write(to_write.encode("ISO-8859-1"))
            self.proc.stdin.flush()
            size_info_str = self.outpipe_fp.read(8).encode(encoding="ISO-8859-1")

        # java "long" is 8 bytes, which python struct calls "long long".
        # java default byte ordering is big-endian.
        size_info = struct.unpack('>Q', size_info_str)[0]
        # print "size expected", size_info

        chunks = []
        curlen = lambda: sum(len(x) for x in chunks)
        while True:
            remaining_size = size_info - curlen()
            if self.comm_mode == 'SOCKET':
                data = sock.recv(remaining_size)
            elif self.comm_mode == 'PIPE':
                data = self.outpipe_fp.read(remaining_size)
            chunks.append(data)
            if curlen() >= size_info: break
            if len(chunks) > 1000:
                LOG.warning("Incomplete value from server")
                return None
            time.sleep(0.01)
        return ''.join(chunks)


def test_modes():
    import pytest
    gosimple(comm_mode='SOCKET')
    gosimple(comm_mode='PIPE')
    with pytest.raises(AssertionError):
        gosimple(comm_mode=None)
    with pytest.raises(AssertionError):
        gosimple(comm_mode='asdfasdf')

def test_coref():
    assert_no_java("no java when starting")
    p = CoreNLP("coref")
    ret = p.parse_doc("I saw Fred. He saw me.")
    pprint(ret)
    assert 'entities' in ret
    assert isinstance(ret['entities'], list)

def gosimple(**kwargs):
    assert_no_java("no java when starting")

    p = CoreNLP("ssplit", **kwargs)
    ret = p.parse_doc("Hello world.")
    # pprint(ret)
    assert len(ret['sentences']) == 1
    assert u' '.join(ret['sentences'][0]['tokens']) == u"Hello world ."

    p.kill_proc_if_running()
    assert_no_java()

def test_paths():
    import pytest
    with pytest.raises(AssertionError):
        CoreNLP("ssplit", corenlp_jars=["/asdfadsf/asdfasdf"])

def assert_no_java(msg=""):
    ps_output = os.popen("ps wux").readlines()
    javalines = [x for x in ps_output if re.search(r'\bbin/java\b', x)]
    print(''.join(javalines))
    assert len(javalines) == 0, msg

# def test_doctimeout():
#     assert_no_java("no java when starting")
#
#     p = CoreNLP("pos")
#     ret = p.parse_doc(open("allbrown.txt").read(), 0.5)
#     assert ret is None
#     p.kill_proc_if_running()
#     assert_no_java()

if __name__=='__main__':
    import sys
    if sys.argv[1]=='modes':
        for mode,d in MODES_items:
            print("  * `%s`: %s" % (mode, d['description']))
    if sys.argv[1]=='modes_json':
        # import json as stdjson
        # print stdjson.dumps(MODES, indent=4)
        print('"%s"' % json.dumps(MODES).replace('"', r'\"'))



""" Utility functions. """

__author__ = 'smartschat'


def clean_via_pos(tokens, pos):
    """ Clean a list of tokens according to their part-of-speech tags.

    In particular, retain only tokens which do not have the part-of-speech tag
    DT (determiner) or POS (possessive 's').

    Args:
        tokens (list(str)): A list of tokens.
        pos (list(str)): A list of corresponding part-of-speech tags.

    Returns:
        list(str): The list of tokens which do not have part-of-speech tag
        DT or POS.
    """
    return [token for token, pos in zip(tokens, pos)
            if pos not in ["DT", "POS"]]



""" Implements instance extraction and decoding for a mention pair model.

This module implements a mention pair model (Soon et al. 2001) within a
framework that expresses coreference resolution as predicting latent structures,
while performing learning using a latent structured perceptron with
cost-augmented inference.

Hence, the mention pair model is expressed as  a latent graph. In particular,
let m_1, ..., m_n be all mentions in a document. We then consider the graph
with nodes m_1, ..., m_n and arcs (m_2, m_1), (m_3, m_2), (m_3, m_1),
(m_4, m_3), ...

Then each arc gets a label, either "+" (coreferent) or "-" (not coreferent).
As each arc is handled individually, each is called a *substructure*.

To implement the mention pair model, this module contains a decoder and
functions that define the search space for training and testing. The decoder
computes the highest-scoring label for an arc, and additionally returns the
true label. The function that defines the search space for testing extracts
arcs as decribed above. For training, arcs are filtered using a procedure
similar to the procedure decribed in Soon et al. (2001).

Reference:
    - Wee Meng Soon, Hwee Tou Ng, and Daniel Chung Yong Lim. 2001. A machine
      learning approach to coreference resolution of noun phrases.
      *Computational Linguistics*, 27(4):521-544.
      http://www.aclweb.org/anthology/J01-4004
"""


from cort.coreference import perceptrons


__author__ = 'martscsn'


def extract_training_substructures(doc):
    """ Extract the search space for training the mention pair model,

    The mention pair model consists in computing labels "+" (coreferent) and
    "-" (not coreferent) for mention pairs. We can view each such labeled pair
    as a labeled arc in a graph. As each pair is handled individually, each
    arc in this graph is a *substructure*.

    The search space is represented as a nested list of mention pairs. Each
    list in the nested list contains only one pair (the search for the optimal
    substructure only chooses a label).

    For training, the list of pairs is obtained via a heuristic similar to
    Soon et al. (2001). We employ the following algorithm: for every
    j >0, if the mention m_j is in some coreference chain, add all pairs
    (m_j, m_{j-1}), (m_j, m_{j-2}), ..., (m_j, m_i) to the list, where
    m_i is the first mention preceding m_j which is coreferent with m_j.

    Args:
        doc (CoNLLDocument): The document to extract substructures from.

    Returns:
        (list(list(Mention, Mention))): The nested list of mention pairs
        describing the search space for the substructures.
    """
    substructures = []

    # iterate over mentions
    for i, ana in enumerate(doc.system_mentions):
        if not ana.attributes["annotated_set_id"]:
            continue

        # iterate in reversed order over candidate antecedents
        for ante in sorted(doc.system_mentions[1:i], reverse=True):
            substructures.append([(ana, ante)])

            if ana.is_coreferent_with(ante):
                break

    return substructures


def extract_testing_substructures(doc):
    """ Extract the search space for predicting with the mention pair model,

    The mention ranking model consists in computing the optimal antecedent
    for an anaphor, which corresponds to predicting an edge in graph. This
    functions extracts the search space for each such substructure (one
    substructure corresponds to one antecedent decision for an anaphor).

    The search space is represented as a nested list of mention pairs. The
    mention pairs are candidate arcs in the graph. The ith list contains the
    ith mention pair, where we assume the following order:

    (m_2, m_1), (m_3, m_2), (m_3, m_1), (m_4, m_3), ...

    Args:
        doc (CoNLLDocument): The document to extract substructures from.

    Returns:
        (list(list(Mention, Mention))): The nested list of mention pairs
        describing the search space for the substructures.
    """
    substructures = []

    # iterate over mentions
    for i, ana in enumerate(doc.system_mentions):

        # iterate in reversed order over candidate antecedents
        for ante in sorted(doc.system_mentions[1:i], reverse=True):
            substructures.append([(ana, ante)])

    return substructures


class MentionPairsPerceptron(perceptrons.Perceptron):
    """ A perceptron for the mention pair model. """
    def argmax(self, substructure, arc_information):
        """ Decoder for the mention pair model.

        Compute highest-scoring label for a pair and the correct label for a
        pair.

        Args:
            substructure (list((Mention, Mention))): The list of mention pairs
                which define the search space for one substructure. For the
                mention pair model, the list contains only one mention pair.
            arc_information (dict((Mention, Mention),
                                  ((array, array, array), list(int), bool)):
                A mapping of arcs (= mention pairs) to information about these
                arcs. The information consists of the features, the costs for
                the arc (for each label), and whether predicting the arc to be
                coreferent is consistent with the gold annotation). The features
                are divided in three arrays: the first array contains the non-
                numeric features, the second array the numeric features, and the
                third array the values for the numeric features. The features
                are represented as integers via feature hashing.

        Returns:
            A 7-tuple describing the highest-scoring label for the pair, and
            the correct label for the pair. The tuple consists of:

                - **best_arcs** (*list((Mention, Mention))*): the pair under
                 consideration (the list contains only one arc),
                - **best_labels** (*list(str)*): the predicted label of the
                  pair. Is either '+' (coreferent) or '-' (not coreferent).
                - **best_scores** (*list(float)*): the score of the predicted
                  label.
                - **best_cons_arcs** (*list((Mention, Mention))*): the pair
                  under consideration (the list contains only one arc),
                - **best_cons_labels** (*list(str)*): the correct label for the
                  arc. Is either '+' (coreferent) or '-' (not coreferent).
                - **best_cons_scores** (*list(float)*): the score of the
                  correct label.
                - **is_consistent** (*bool*): whether the predicted label is the
                  same as the correct label.
        """
        arc = substructure[0]
        consistent = arc_information[arc][2]

        score_coref = self.score_arc(arc, arc_information, "+")
        score_non_coref = self.score_arc(arc, arc_information, "-")

        if score_coref >= score_non_coref:
            label = "+"
            score = score_coref
        else:
            label = "-"
            score = score_non_coref

        if consistent:
            coref_label = "+"
            coref_score = score_coref
        else:
            coref_label = "-"
            coref_score = score_non_coref

        return ([arc], [label], [score], [arc], [coref_label], [coref_score],
                label == coref_label)

    def get_labels(self):
        """ Get the graph labels employed by the mention pair approach..

        The mention pair approach uses labels '+' (coreferent) and '-' (not
        coreferent).

        Returns:
            list(str): The list ['+', '-'] of graph labels.
        """
        return ["+", "-"]



""" Extract instances and features from a corpus. """

import array
import multiprocessing
import sys

import mmh3
import numpy


__author__ = 'martscsn'


# for python 2 multiprocessing
def unwrap_extract_doc(arg, **kwarg):
    return InstanceExtractor._extract_doc(*arg, **kwarg)


class InstanceExtractor:
    """ Extract instances and their corresponding features from a corpus.

    Attributes:
        extract_substructures (function: CoNLLDocument ->
             list(list((Mention,Mention)))): Function for extracting the search
             space for a coreference resolution approach. The ith list in the
             nested list contains the search space for the ith substructure.
             The search space is represented as a nested list of mention pairs,
             which are candidate arcs in the graph to predict.
        mention_features (list(function: Mention -> str)): A list of features
            for mentions.
        pairwise_features (list(function: (Mention, Mention) -> str)): A list
            of features for mention pairs.
        cost_function (function: (Mention, Mention) -> int): A function
            assigning costs to mention pairs.
        labels (list(str)): A list of arc labels emplyoed by the approach.
            Defaults to the list containing only "+".
    """
    def __init__(self,
                 extract_substructures,
                 mention_features,
                 pairwise_features,
                 cost_function,
                 labels=("+",)):
        """ Initialize instance and feature extraction.

        Args:
            extract_substructures (function: CoNLLDocument ->
                list(list((Mention,Mention)))): Function for extracting the
                search space for a coreference resolution approach. The ith
                list in the nested list contains the search space for the ith
                substructure. The search space is represented as a nested list
                of mention pairs, which are candidate arcs in the graph to
                predict.
            mention_features (list(function: Mention -> str)): A list of
                features for mentions.
            pairwise_features (list(function: (Mention, Mention) -> str)): A
                list of features for mention pairs.
            cost_function (function: (Mention, Mention) -> int): A function
                assigning costs to mention pairs.
            labels (list(str)): A list of arc labels emplyoed by the
                approach.
        """
        self.extract_substructures = extract_substructures
        self.mention_features = mention_features
        self.pairwise_features = pairwise_features
        self.cost_function = cost_function
        self.labels = labels

    def extract(self, corpus):
        """ Extract instances and features from a corpus.

        Args:
            corpus (Corpus): The corpus to extract instances and features from.

        Returns:
            A tuple which describes the extracted instances and their
            features. The individual components are:

            * substructures (list(list((Mention, Mention)))): The search space
                for the substructures, defined by a nested list. The ith list
                contains the search space for the ith substructure.
            * arc_information (dict((Mention, Mention),
                                    ((array, array, array), list(int), bool)):
                A mapping of arcs (= mention pairs) to information about these
                arcs. The information consists of the features, the costs for
                the arc (for each label), and whether predicting the arc to be
                coreferent is consistent with the gold annotation). The features
                are divided in three arrays: the first array contains the non-
                numeric features, the second array the numeric features, and the
                third array the values for the numeric features. The features
                are represented as integers via feature hashing.
        """

        all_substructures = []
        arc_information = {}

        id_to_doc_mapping = {}
        for doc in corpus:
            id_to_doc_mapping[doc.identifier] = doc

        pool = multiprocessing.Pool(maxtasksperchild=1)

        if sys.version_info[0] == 2:
            results = pool.map(unwrap_extract_doc,
                               zip([self] * len(corpus.documents),
                                   corpus.documents))
        else:
            results = pool.map(self._extract_doc, corpus.documents)

        pool.close()
        pool.join()

        num_labels = len(self.labels)

        for result in results:
            (doc_identifier,
             anaphors,
             antecedents,
             nonnumeric_features,
             numeric_features,
             numeric_vals,
             costs,
             consistency,
             nonnumeric_feature_mapping,
             numeric_feature_mapping,
             substructures_mapping) = result

            doc = id_to_doc_mapping[doc_identifier]

            for i in range(0, len(substructures_mapping) - 1):
                struct = []
                begin = substructures_mapping[i]
                end = substructures_mapping[i + 1]

                for pair_index in range(begin, end):
                    arc = (doc.system_mentions[anaphors[pair_index]],
                           doc.system_mentions[antecedents[pair_index]])

                    struct.append(arc)

                    # find position of arc's features in document array
                    nonnumeric_features_start = nonnumeric_feature_mapping[
                        pair_index]
                    nonnumeric_features_end = nonnumeric_feature_mapping[
                        pair_index + 1]

                    numeric_features_start = numeric_feature_mapping[pair_index]
                    numeric_features_end = numeric_feature_mapping[
                        pair_index + 1]

                    arc_information[arc] = \
                        ((nonnumeric_features[
                          nonnumeric_features_start:nonnumeric_features_end
                          ],
                          numeric_features[
                          numeric_features_start:numeric_features_end
                          ],
                          numeric_vals[
                          numeric_features_start:numeric_features_end
                          ]),
                         costs[
                         num_labels * pair_index:num_labels * pair_index
                         + num_labels],
                         consistency[pair_index])

                all_substructures.append(struct)

        # in python 2, array.array does not support the buffer interface
        if sys.version_info[0] == 2:
            for arc in arc_information:
                feats, cost, cons = arc_information[arc]
                arc_information[arc] = (
                    (numpy.array(feats[0], dtype=numpy.uint32),
                     numpy.array(feats[1], dtype=numpy.uint32),
                     numpy.array(feats[2], dtype="float32")),
                    numpy.array(cost, dtype=float),
                    cons)

        return all_substructures, arc_information

    def _extract_doc(self, doc):
        cache = {}
        substructures = self.extract_substructures(doc)

        mentions_to_ids = {}

        for i, mention in enumerate(doc.system_mentions):
            mentions_to_ids[mention] = i

        anaphors = array.array('H')
        antecedents = array.array('H')
        costs = array.array('H')
        consistency = array.array('B')
        nonnumeric_feature_mapping = array.array('I')
        numeric_feature_mapping = array.array('I')
        substructures_mapping = array.array('I')
        nonnumeric_features = array.array('I')
        numeric_features = array.array('I')
        numeric_vals = array.array("f")

        nonnumeric_feature_mapping.append(0)
        numeric_feature_mapping.append(0)
        substructures_mapping.append(0)

        for struct in substructures:
            # skip empty
            if not struct:
                continue

            for arc in struct:
                # ids for anaphor and antecedent
                anaphors.append(mentions_to_ids[arc[0]])
                antecedents.append(mentions_to_ids[arc[1]])

                # cost for each label
                for label in self.labels:
                    costs.append(self.cost_function(arc, label))

                # is decision to make them coreferent consistent with gold?
                consistency.append(arc[0].decision_is_consistent(arc[1]))

                # features for the arc: stored in array which applies to whole
                # document
                (arc_nonnumeric_features, arc_numeric_features,
                 arc_numeric_vals) = self._extract_features(arc, cache)

                nonnumeric_features.extend(arc_nonnumeric_features)
                numeric_features.extend(arc_numeric_features)
                numeric_vals.extend(arc_numeric_vals)

                # auxiliary arrays that store the position of features for arcs
                # in the document array
                nonnumeric_feature_mapping.append(
                    nonnumeric_feature_mapping[-1] + len(
                        arc_nonnumeric_features))
                numeric_feature_mapping.append(
                    numeric_feature_mapping[-1] + len(arc_numeric_features))

            # store position of substructures in document array
            substructures_mapping.append(substructures_mapping[-1] +
                                         len(struct))

        return (doc.identifier,
                anaphors,
                antecedents,
                nonnumeric_features,
                numeric_features,
                numeric_vals,
                costs,
                consistency,
                nonnumeric_feature_mapping,
                numeric_feature_mapping,
                substructures_mapping)

    def _extract_features(self, arc, cache):
        anaphor, antecedent = arc
        inst_feats = []
        numeric_features = []

        numeric_types = {"float", "int"}

        if not antecedent.is_dummy():
            # mention features
            for mention in [anaphor, antecedent]:
                if mention not in cache:
                    cache[mention] = [feature(mention) for feature
                                      in self.mention_features]

            ana_features = cache[anaphor]
            ante_features = cache[antecedent]

            # first: non-numeric features (categorial, boolean)
            inst_feats += ["ana_" + feat + "=" + str(val) for feat, val in
                           ana_features if type(val).__name__ not in
                           numeric_types]

            len_ana_features = len(inst_feats)

            inst_feats += ["ante_" + feat + "=" + str(val) for feat, val in
                           ante_features if type(val).__name__ not in
                           numeric_types]

            # concatenated features
            inst_feats += ["ana_" + ana_info[0] + "=" + str(ana_info[1]) +
                           "^ante_" + ante_info[0] + "=" + str(ante_info[1])
                           for ana_info, ante_info in
                           zip(ana_features, ante_features)]

            # pairwise features
            pairwise_features = [feature(anaphor, antecedent) for feature
                                 in self.pairwise_features]
            inst_feats += [feature + "=" + str(val) for feature, val
                           in pairwise_features
                           if val and type(val).__name__ not in numeric_types]

            # feature combinations
            fine_type_indices = {len_ana_features * i for i
                                 in [0, 1, 2]}

            inst_feats += [
                inst_feats[i] + "^" + word for i in fine_type_indices
                for j, word in enumerate(inst_feats)
                if j not in fine_type_indices
            ]

            # now numeric features
            ana_numeric = [("ana_" + feat, val) for feat, val
                           in ana_features
                           if type(val).__name__ in numeric_types]
            ante_numeric = [("ante_" + feat, val) for feat, val
                            in ante_features
                            if type(val).__name__ in numeric_types]
            pair_numeric = [(feat, val) for feat, val in pairwise_features
                            if type(val).__name__ in numeric_types]

            # feature combinations for numeric features
            for numeric_features in [ana_numeric, ante_numeric, pair_numeric]:
                numeric_features += [
                    (inst_feats[i] + "^" + numeric_features[j][0],
                     numeric_features[j][1]) for i in fine_type_indices
                    for j, numeric_feature in enumerate(numeric_features)
                ]

            numeric_features = ana_numeric + ante_numeric + pair_numeric

        # to hash
        all_nonnumeric_feats = array.array(
            'I', [mmh3.hash(word.encode("utf-8")) & 2 ** 24 - 1 for word
                  in inst_feats])
        all_numeric_feats = array.array(
            'I', [mmh3.hash(word.encode("utf-8")) & 2 ** 24 - 1 for word, _
                  in numeric_features])
        numeric_vals = array.array("f", [val for _, val in numeric_features])

        return all_nonnumeric_feats, all_numeric_feats, numeric_vals
